import pytest

from os.path import exists
import tempfile

import numpy as np
import reciprocalspaceship as rs
import torch

from scipy.stats import pearsonr
from SFC_Torch.io import PDBParser
from SFC_Torch.Fmodel import SFcalculator
from SFC_Torch.utils import assert_numpy, assert_tensor

@pytest.mark.parametrize("case", [1, 2, 3])
def test_constructor_SFcalculator(data_pdb, data_mtz_exp, data_cif, data_sfcif, case):
    if case == 1:
        sfcalculator = SFcalculator(
            data_pdb, mtzdata=data_mtz_exp, set_experiment=True)
        sfcalculator.inspect_data()
        bins_labels = sfcalculator.assign_resolution_bins(return_labels=True)
        assert sfcalculator.inspected
        assert np.isclose(assert_numpy(sfcalculator.solventpct), 0.1667, 1e-3)
        assert sfcalculator.gridsize == [48, 60, 60]
        assert len(sfcalculator.HKL_array) == 3197
        assert len(sfcalculator.Hasu_array) == 3255
        assert len(sfcalculator.bins) == 3197
        assert np.all(np.sort(np.unique(sfcalculator.bins)) == np.arange(0,10))
        assert len(bins_labels) == 10
    elif case == 2:
        sfcalculator = SFcalculator(
            data_cif, mtzdata=data_sfcif, set_experiment=True)
        sfcalculator.inspect_data()
        bins_labels = sfcalculator.assign_resolution_bins(return_labels=True)
        assert sfcalculator.inspected
        assert np.isclose(assert_numpy(sfcalculator.solventpct), 0.1667, 1e-3)
        assert sfcalculator.gridsize == [50, 64, 64]
        assert len(sfcalculator.HKL_array) == 3256
        assert len(sfcalculator.Hasu_array) == 4035
        assert len(sfcalculator.bins) == 3256
        assert np.all(np.sort(np.unique(sfcalculator.bins)) == np.arange(0,10))
        assert len(bins_labels) == 10
    elif case == 3:
        sfcalculator = SFcalculator(
            data_pdb, mtzdata=None, dmin=2.5, set_experiment=True)
        sfcalculator.inspect_data()
        assert sfcalculator.inspected
        assert np.isclose(assert_numpy(sfcalculator.solventpct), 0.1667, 1e-3)
        assert sfcalculator.gridsize == [40, 48, 48]
        assert sfcalculator.HKL_array is None
        assert len(sfcalculator.Hasu_array) == 1747
    assert len(sfcalculator.atom_name) == 488


def test_constructor_SFcalculator_obj(data_pdb, data_mtz_exp):
    pdbmodel = PDBParser(data_pdb)
    mtzdata = rs.read_mtz(data_mtz_exp)
    sfcalculator = SFcalculator(
        pdbmodel, mtzdata=mtzdata, set_experiment=True)
    sfcalculator.inspect_data()
    bins_labels = sfcalculator.assign_resolution_bins(return_labels=True)
    assert sfcalculator.inspected
    assert np.isclose(assert_numpy(sfcalculator.solventpct), 0.1667, 1e-3)
    assert sfcalculator.gridsize == [48, 60, 60]
    assert len(sfcalculator.HKL_array) == 3197
    assert len(sfcalculator.Hasu_array) == 3255
    assert len(sfcalculator.bins) == 3197
    assert np.all(np.sort(np.unique(sfcalculator.bins)) == np.arange(0,10))
    assert len(bins_labels) == 10

@pytest.mark.parametrize("Return", [True, False])
@pytest.mark.parametrize("Anomalous", [True, False])
def test_calc_fall(data_pdb, data_mtz_exp, data_mtz_fmodel_ksol0, data_mtz_fmodel_ksol1, Return, Anomalous):
    sfcalculator = SFcalculator(
        data_pdb, mtzdata=data_mtz_exp, set_experiment=True, anomalous=Anomalous)
    sfcalculator.inspect_data()
    Fprotein = sfcalculator.calc_fprotein(Return=Return)
    Fsolvent = sfcalculator.calc_fsolvent(
        dmin_mask=6.0, dmin_nonzero=3.0, Return=Return)
    sfcalculator.get_scales_lbfgs(ls_steps=2, r_steps=2, ls_lr=0.00001, r_lr=0.00001, verbose=False)
    Ftotal = sfcalculator.calc_ftotal()
    assert len(Ftotal) == 3197
    assert assert_numpy(sfcalculator.r_free) < 0.15
    sfcalculator.get_scales_adam(lr=0.01, n_steps=20, sub_ratio=0.3, initialize=True, verbose=False)
    assert assert_numpy(sfcalculator.r_free) < 0.15

    Fcalc = rs.read_mtz(data_mtz_fmodel_ksol0)
    Fmodel = rs.read_mtz(data_mtz_fmodel_ksol1)
    assert (Fmodel.get_hkls() == sfcalculator.HKL_array).all()

    calc_mag = Fcalc['FMODEL'].to_numpy()
    calc_ph = np.deg2rad(Fcalc['PHIFMODEL'].to_numpy())
    Fcalc_complex = np.array([complex(mag*np.cos(ph), mag*np.sin(ph))
                              for mag, ph in zip(calc_mag, calc_ph)])
    model_mag = Fmodel['FMODEL'].to_numpy()
    model_ph = np.deg2rad(Fmodel['PHIFMODEL'].to_numpy())
    Fmodel_complex = np.array([complex(mag*np.cos(ph), mag*np.sin(ph))
                               for mag, ph in zip(model_mag, model_ph)])

    Fmask_complex = Fmodel_complex - Fcalc_complex

    if Return:
        Fsolvent_arr = assert_numpy(Fsolvent)
        Fprotein_arr = assert_numpy(Fprotein)
        try:
            assert pearsonr(np.abs(Fprotein_arr),
                            Fcalc['FMODEL'].to_numpy()).statistic > 0.99
            assert pearsonr(np.abs(Fsolvent_arr),
                            np.abs(Fmask_complex)).statistic > 0.84
        except:
            assert pearsonr(np.abs(Fprotein_arr),
                            Fcalc['FMODEL'].to_numpy())[0] > 0.99
            assert pearsonr(np.abs(Fsolvent_arr),
                            np.abs(Fmask_complex))[0] > 0.84
    else:
        assert Fsolvent is None
        assert Fprotein is None
        Fprotein_arr = assert_numpy(sfcalculator.Fprotein_HKL)
        Fsolvent_arr = assert_numpy(sfcalculator.Fmask_HKL)
        try:
            assert pearsonr(np.abs(Fprotein_arr),
                            Fcalc['FMODEL'].to_numpy()).statistic > 0.99
            assert pearsonr(np.abs(Fsolvent_arr),
                            np.abs(Fmask_complex)).statistic > 0.84
        except:
            assert pearsonr(np.abs(Fprotein_arr),
                            Fcalc['FMODEL'].to_numpy())[0] > 0.99
            assert pearsonr(np.abs(Fsolvent_arr),
                            np.abs(Fmask_complex))[0] > 0.84

def test_calc_fall_setter(data_pdb, data_mtz_exp, Return=True):
    sfcalculator = SFcalculator(
        data_pdb, mtzdata=data_mtz_exp, set_experiment=True)
    pdbmodel = PDBParser(data_pdb)
    sfcalculator.inspect_data()
    Fprotein = sfcalculator.calc_fprotein(
        atoms_position_tensor=assert_tensor(pdbmodel.atom_pos, arr_type=torch.float32),
        atoms_biso_tensor=assert_tensor(pdbmodel.atom_b_iso, arr_type=torch.float32),
        atoms_aniso_uw_tensor=assert_tensor(pdbmodel.atom_b_aniso, arr_type=torch.float32),
        atoms_occ_tensor=assert_tensor(pdbmodel.atom_occ, arr_type=torch.float32),
        Return=Return
    )
    Fsolvent = sfcalculator.calc_fsolvent(
        dmin_mask=6.0, dmin_nonzero=3.0, Return=Return
    )
    sfcalculator.get_scales_lbfgs(ls_steps=2, r_steps=2, ls_lr=0.00001, r_lr=0.00001, verbose=False)
    Ftotal = sfcalculator.calc_ftotal()
    assert len(Ftotal) == 3197
    assert assert_numpy(sfcalculator.r_free) < 0.15
    sfcalculator.get_scales_adam(lr=0.01, n_steps=20, sub_ratio=0.3, initialize=True, verbose=False)
    assert assert_numpy(sfcalculator.r_free) < 0.15

def test_calc_ftotal_nodata(data_pdb):
    sfcalculator = SFcalculator(
        data_pdb, mtzdata=None, dmin=2.5, set_experiment=False)
    sfcalculator.inspect_data()
    sfcalculator.calc_fprotein(Return=False)
    sfcalculator.calc_fsolvent(
        dmin_mask=6.0, dmin_nonzero=3.0, Return=False)
    sfcalculator.init_scales(requires_grad=False)
    Ftotal = sfcalculator.calc_ftotal()
    assert len(Ftotal) == 1747


@pytest.mark.parametrize("partition_size", [1, 3, 5])
@pytest.mark.parametrize("Anomalous", [True, False])
def test_calc_fall_batch(data_pdb, data_mtz_exp, Anomalous, partition_size):
    sfcalculator = SFcalculator(
        data_pdb, mtzdata=data_mtz_exp, set_experiment=True, anomalous=Anomalous)
    sfcalculator.inspect_data()
    atoms_pos_batch = torch.tile(sfcalculator.atom_pos_orth, [5, 1, 1])

    Fprotein = sfcalculator.calc_fprotein(Return=True)
    Fsolvent = sfcalculator.calc_fsolvent(
        dmin_mask=6.0, dmin_nonzero=3.0, Return=True)
    Fprotein_batch = sfcalculator.calc_fprotein_batch(
        atoms_pos_batch, Return=True, PARTITION=partition_size)
    Fsolvent_batch = sfcalculator.calc_fsolvent_batch(
        dmin_mask=6.0, dmin_nonzero=3.0, Return=True, PARTITION=partition_size)
    
    sfcalculator.init_scales(requires_grad=False)
    Ftotal = sfcalculator.calc_ftotal()
    Ftotal_batch = sfcalculator.calc_ftotal_batch()

    assert len(Fprotein_batch) == 5
    assert np.all(np.isclose(assert_numpy(Fprotein_batch[3]),
                             assert_numpy(Fprotein),
                             rtol=1e-5, atol=5e-3))
    assert np.all(np.isclose(assert_numpy(sfcalculator.Fprotein_asu_batch[4]),
                             assert_numpy(sfcalculator.Fprotein_asu),
                             rtol=1e-5, atol=5e-3))

    assert len(Fsolvent_batch) == 5
    assert np.all(np.isclose(assert_numpy(torch.abs(Fsolvent)),
                             assert_numpy(torch.abs(Fsolvent_batch[1])),
                             rtol=1e-3, atol=1e-2))

    assert np.all(np.isclose(assert_numpy(torch.abs(Ftotal_batch[2])),
                             assert_numpy(torch.abs(Ftotal)),
                             rtol=1e-3, atol=1e1))


def test_prepare_dataset(data_pdb, data_mtz_exp):
    sfcalculator = SFcalculator(
        data_pdb, mtzdata=data_mtz_exp, set_experiment=True)
    sfcalculator.inspect_data()
    sfcalculator.calc_fprotein(Return=False)
    sfcalculator.calc_fsolvent(
        dmin_mask=6.0, dmin_nonzero=3.0, Return=False)
    sfcalculator.init_scales(requires_grad=False)
    sfcalculator.calc_ftotal()
    ds = sfcalculator.prepare_dataset("HKL_array", "Ftotal_HKL")
    assert len(ds) == 3197

def test_savePDB(data_pdb, data_mtz_exp):
    sfcalculator = SFcalculator(
        data_pdb, mtzdata=data_mtz_exp, set_experiment=True)
    sfcalculator.inspect_data()
    sfcalculator.calc_fprotein(Return=False)
    sfcalculator.calc_fsolvent(
        dmin_mask=6.0, dmin_nonzero=3.0, Return=False)
    sfcalculator.init_scales(requires_grad=False)
    sfcalculator.calc_ftotal()
    with tempfile.NamedTemporaryFile(suffix=".pdb") as temp:
        sfcalculator.savePDB(temp.name)
        assert exists(temp.name)
    



import pytest

import numpy as np
import reciprocalspaceship as rs
import torch

from SFC_Torch.Fmodel import SFcalculator
from SFC_Torch.utils import try_gpu, vdw_rad_tensor, unitcell_grid_center
from SFC_Torch.voxel import voxelvalue_torch_p1, voxelvalue_torch_p1_savememory
from SFC_Torch.voxel import voxel_1dto3d_np, voxel_1dto3d_torch


@pytest.mark.parametrize("binary", [True, False])
def test_voxelvalue_torch_p1_sm(data_pdb, data_mtz_exp, binary):
    sfcalculator = SFcalculator(
        data_pdb, mtzdata=data_mtz_exp, set_experiment=True)
    vdw_rad = vdw_rad_tensor(sfcalculator.atom_name)
    uc_grid_orth_tensor = unitcell_grid_center(sfcalculator.unit_cell,
                                               spacing=4.5, return_tensor=True)
    CUTOFF = 0.0001
    N_grid = len(uc_grid_orth_tensor[:, 0])
    spacing = torch.max(uc_grid_orth_tensor[1] - uc_grid_orth_tensor[0])
    # s ~ log(1/c -1) / (d/2 - r)
    spacing = torch.maximum(spacing, torch.tensor(3.5))
    steepness = np.log(1.0/CUTOFF - 1.0)/(spacing/2.0 - 1.5)

    voxel_map_p1 = voxelvalue_torch_p1(uc_grid_orth_tensor, sfcalculator.atom_pos_orth,
                                       sfcalculator.unit_cell, sfcalculator.space_group, vdw_rad,
                                       s=steepness, binary=binary, cutoff=CUTOFF)
    voxel_map_p1_sm = voxelvalue_torch_p1_savememory(uc_grid_orth_tensor, sfcalculator.atom_pos_orth,
                                                     sfcalculator.unit_cell, sfcalculator.space_group, vdw_rad,
                                                     s=steepness, binary=binary, cutoff=CUTOFF)

    if binary:
        assert np.all(voxel_map_p1.cpu().numpy() ==
                      voxel_map_p1_sm.cpu().numpy())
    else:
        assert np.all(np.isclose(voxel_map_p1.cpu().numpy(),
                      voxel_map_p1_sm.cpu().numpy()))


def test_voxel_1dto3d(data_pdb, data_mtz_exp):
    sfcalculator = SFcalculator(
        data_pdb, mtzdata=data_mtz_exp, set_experiment=True)
    vdw_rad = vdw_rad_tensor(sfcalculator.atom_name)
    uc_grid_orth_tensor = unitcell_grid_center(sfcalculator.unit_cell,
                                               spacing=4.5, return_tensor=True)
    CUTOFF = 0.0001
    N_grid = len(uc_grid_orth_tensor[:, 0])
    spacing = torch.max(uc_grid_orth_tensor[1] - uc_grid_orth_tensor[0])
    # s ~ log(1/c -1) / (d/2 - r)
    spacing = torch.maximum(spacing, torch.tensor(3.5))
    steepness = np.log(1.0/CUTOFF - 1.0)/(spacing/2.0 - 1.5)

    voxel_map_p1 = voxelvalue_torch_p1(uc_grid_orth_tensor, sfcalculator.atom_pos_orth,
                                       sfcalculator.unit_cell, sfcalculator.space_group, vdw_rad,
                                       s=steepness, binary=True, cutoff=CUTOFF)
    a, b, c, _, _, _ = sfcalculator.unit_cell.parameters
    na = int(a/4.5)
    nb = int(b/4.5)
    nc = int(c/4.5)
    map_1 = voxel_1dto3d_np(voxel_map_p1.cpu().numpy(), na, nb, nc)
    map_2 = voxel_1dto3d_torch(voxel_map_p1, na, nb, nc).cpu().numpy()
    assert np.all(map_1 == map_2)



import pytest
from os.path import dirname, abspath, join

@pytest.fixture
def data_pdb():
    datapath = "data/1dur.pdb"
    filename = abspath(join(dirname(__file__), datapath))
    return filename

@pytest.fixture
def data_cif():
    datapath = "data/1dur.cif"
    filename = abspath(join(dirname(__file__), datapath))
    return filename

@pytest.fixture
def data_sfcif():
    datapath = "data/1dur-sf.cif"
    filename = abspath(join(dirname(__file__), datapath))
    return filename

@pytest.fixture
def data_mtz_exp():
    datapath = "data/1dur.mtz"
    filename = abspath(join(dirname(__file__), datapath))
    return filename

@pytest.fixture
def data_mtz_fmodel_ksol0():
    datapath = "data/1dur_vanilla.mtz"
    filename = abspath(join(dirname(__file__), datapath))
    return filename

@pytest.fixture
def data_mtz_fmodel_ksol1():
    datapath = "data/1dur_ksol1.mtz"
    filename = abspath(join(dirname(__file__), datapath))
    return filename




"""
Calculate Structural Factor from an atomic model: F_model = k_total * (F_calc + k_mask * F_mask)

Note:
1. We use direct summation for the F_calc
2. We support anomalous scattering with cromer_liberman equation
3. We use a bulk solvent masking 

Written in PyTorch
"""
from __future__ import annotations

from typing import Optional, List
import gemmi
import time
import numpy as np
import torch
import reciprocalspaceship as rs

from .symmetry import generate_reciprocal_asu, expand_to_p1
from .mask import reciprocal_grid, rsgrid2realmask, realmask2Fmask
from .utils import try_gpu, DWF_aniso, DWF_iso, diff_array, asu2HKL, aniso_scaling
from .utils import vdw_rad_tensor, unitcell_grid_center, bin_by_logarithmic
from .utils import r_factor, assert_numpy, assert_tensor
from .packingscore import packingscore_voxelgrid_torch
from .io import PDBParser

class SFcalculator(object):
    """
    A class to formalize the structural factor calculation.
    """

    def __init__(
        self,
        pdbmodel:       str | PDBParser,
        mtzdata:        str | rs.DataSet = None,
        n_bins:         int = 10, 
        dmin:           Optional[float] = None,
        anomalous:      bool = False,
        wavelength:     Optional[float] = None,
        set_experiment: bool = True,
        expcolumns:     List[str] = ["FP", "SIGFP"],
        freeflag:       str = "FreeR_flag",
        testset_value:  int = 0,
        device:         torch.device = try_gpu(),
        mode:           str = "xray",
    ) -> None:
        """
        Initialize with necessary reusable information, like spacegroup, unit cell info, HKL_list, et.c.

        Parameters:
        -----------
        pdbmodel: str | PDBParser
            path to the PDB file or an instance of PDBparser. Will use its unit cell info, space group info, atom name info,
            atom position info, atoms B-factor info and atoms occupancy info to initialize the instance.

        mtzdata: str | rs.Dataset 
            path to the mtz data or instance of rs.Dataset, will use the HKL list in the mtz instead, override dmin with an inference
            also support reflection data in cif file
    
        n_bins: str, default 10
            Number of resolution bins used in the reciprocal space
        
        dmin: float | None
            highest resolution in the map in Angstrom, to generate Miller indices in recirpocal ASU

        anomalous: bool, default False
            Whether or not to include anomalous scattering in the calculation

        wavelength: float | None
            The wavelength of scattering source in A

        set_experiment: bool, default True
            Whether or not to set Fo, SigF, free_flag and Outlier from the experimental mtz file. It only works when
            the mtzfile_dir is not None

        expcolumns: List[str], default ['FP', 'SIGFP']
            list of column names used as expeimental data

        freeflag: str, default "FreeR_flag"
            column names used as free flag, default "FreeR_flag". Also could be "FREE" and "R-free-flags" 
            in Phenix(CNS/XPOLAR) convention
        
        testset_value: int, default 0
            Test set values in the freeflag column

        device: torch.device

        mode: str, default xray
            xray or cryoem, use xray atomic factors or elelctron atomic factors
        """
        
        self.wavelength = wavelength
        self.anomalous = anomalous
        self.device = device
        assert mode in ["xray", "cryoem"], ValueError(
                "mode has to be xray or cryoem!"
            )
        self.mode = mode
        self.init_pdb(pdbmodel)
        if mtzdata is not None:
            self.init_mtz(mtzdata, n_bins, expcolumns, set_experiment, freeflag, testset_value, dmin)
        else:
            self.init_withoutmtz(dmin, n_bins)
        self._init_spacegroup()
        self._init_cell()
        self.init_atomic_scattering()
        self.inspected = False

    def init_pdb(self, pdbmodel: str | PDBParser):
        """
        set pdb topology, symmetry operations, unit_cell properties, and initialize model coordinates
        """
        if isinstance(pdbmodel, str):
            self._pdb = PDBParser(pdbmodel)  # sfc.PDBparser object
        elif isinstance(pdbmodel, PDBParser):
            self._pdb = pdbmodel
        else:
            raise TypeError("pdbmodel should be PDBparser instance or path str to a pdb file!")
        
        # set molecule related property
        # Tensor atom's Positions in orthogonal space, [Nc,3]
        self._atom_pos_orth = assert_tensor(self._pdb.atom_pos, device=self.device, arr_type=torch.float32)
        # Tensor of anisotropic B Factor in matrix form, [Nc,3,3]
        self._atom_aniso_uw = assert_tensor(self._pdb.atom_b_aniso, device=self.device, arr_type=torch.float32)
        # Tensor of isotropic B Factor [B1,B2,...], [Nc]
        self._atom_b_iso = assert_tensor(self._pdb.atom_b_iso, device=self.device, arr_type=torch.float32)
        # Tensor of occupancy [P1,P2,....], [Nc]
        self._atom_occ = assert_tensor(self._pdb.atom_occ, device=self.device, arr_type=torch.float32)

        if self.anomalous:
            # Try to get the wavelength from PDB remarks
            try:
                line_index = np.argwhere(
                    ["WAVELENGTH OR RANGE" in i for i in self._pdb.pdb_header]
                )
                pdb_wavelength = eval(
                    self._pdb.pdb_header[line_index[0, 0]].split()[-1]
                )
                if self.wavelength is not None:
                    assert np.isclose(pdb_wavelength, self.wavelength, atol=0.05)
                else:
                    self.wavelength = pdb_wavelength
            except:
                print(
                    "Can't find wavelength record in the PDB file, or it doesn't match your input wavelength!"
                )

    @property
    def space_group(self):
        return self._pdb.spacegroup
    
    @space_group.setter
    def space_group(self, spacegroup):
        self._pdb.set_spacegroup(spacegroup)
        self._init_spacegroup()
    
    @property
    def unit_cell(self):
        return self._pdb.cell
    
    @unit_cell.setter
    def unit_cell(self, cell):
        self._pdb.set_unitcell(cell)
        self._init_cell()

    def _init_spacegroup(self):
        # Set up spacegroup related property
        self.operations = self.space_group.operations()  # gemmi.GroupOps object
        self.R_G_tensor_stack = assert_tensor(
            np.array([np.array(sym_op.rot) / sym_op.DEN for sym_op in self.operations]),
            arr_type=torch.float32,
            device=self.device,
        )
        self.T_G_tensor_stack = assert_tensor(
            np.array(
                [np.array(sym_op.tran) / sym_op.DEN for sym_op in self.operations]
            ),
            arr_type=torch.float32,
            device=self.device,
        )
    
    def _init_cell(self):
        # Set up unit cell related property
        self.orth2frac_tensor = torch.tensor(
            self.unit_cell.fractionalization_matrix.tolist(), device=self.device
        ).type(torch.float32)
        self.frac2orth_tensor = torch.tensor(
            self.unit_cell.orthogonalization_matrix.tolist(), device=self.device
        ).type(torch.float32)
        self.reciprocal_cell = self.unit_cell.reciprocal()  # gemmi.UnitCell object
        # [ar, br, cr, cos(alpha_r), cos(beta_r), cos(gamma_r)]
        self.reciprocal_cell_paras = torch.tensor(
            [
                self.reciprocal_cell.a,
                self.reciprocal_cell.b,
                self.reciprocal_cell.c,
                np.cos(np.deg2rad(self.reciprocal_cell.alpha)),
                np.cos(np.deg2rad(self.reciprocal_cell.beta)),
                np.cos(np.deg2rad(self.reciprocal_cell.gamma)),
            ],
            device=self.device,
        ).type(torch.float32)
    
    @property
    def atom_pos_orth(self):
        return self._atom_pos_orth
    
    @atom_pos_orth.setter
    def atom_pos_orth(self, atoms_positions_tensor: torch.Tensor):
        assert (
            atoms_positions_tensor.shape == torch.Size([self.n_atoms, 3])
        ), f"atoms_postions_tensor should be in shape [{self.n_atoms}, 3]!"
        self._atom_pos_orth = atoms_positions_tensor
    
    @property
    def atom_b_iso(self):
        return self._atom_b_iso
    
    @atom_b_iso.setter
    def atom_b_iso(self, atoms_biso_tensor: torch.Tensor):
        assert atoms_biso_tensor.shape == torch.Size([self.n_atoms]), f"atoms_biso_tensor should be in shape [{self.n_atoms}]"
        self._atom_b_iso = atoms_biso_tensor
    
    @property
    def atom_aniso_uw(self):
        return self._atom_aniso_uw
    
    @atom_aniso_uw.setter
    def atom_aniso_uw(self, atoms_aniso_uw_tensor: torch.Tensor):
        assert (
            atoms_aniso_uw_tensor.shape == torch.Size([self.n_atoms, 3, 3]) 
        ), f"atoms_aniso_uw_tensor should be in shape [{self.n_atoms}, 3, 3]"
        self._atom_aniso_uw = atoms_aniso_uw_tensor

    @property
    def atom_occ(self):
        return self._atom_occ
    
    @atom_occ.setter
    def atom_occ(self, atoms_occ_tensor: torch.Tensor):
        assert atoms_occ_tensor.shape == torch.Size([self.n_atoms]), f"atoms_occ_tensor should be in shape [{self.n_atoms}]"
        self._atom_occ = atoms_occ_tensor        

    @property
    def atom_pos_frac(self):
        """
        Tensor of atom's Positions in fractional space, [Nc,3]
        """
        return self.orth2frac(self.atom_pos_orth)

    @property
    def cra_name(self):
        """
        A list of Chain-Residue-Atom name, ['A-0-ALA-CA', ...]
        """
        return self._pdb.cra_name
    
    @property
    def atom_name(self):
        """
        A list of element name, ['C', 'N', 'H', ...]
        """
        return self._pdb.atom_name
    
    @property
    def n_atoms(self):
        return len(self.atom_name)
    
    @property
    def unique_atom(self):
        return list(set(self.atom_name))
    
    def frac2orth(self, frac_pos: torch.Tensor) -> torch.Tensor:
        """
        Convert fractional coordinates to orthogonal coordinates

        Args:
            frac_pos: torch.Tensor, [n_points, ..., 3]
        
        Returns:
            orthogonal coordinates, torch.Tensor
        """
        orth_pos = torch.einsum("n...x,yx->n...y", frac_pos, self.frac2orth_tensor)
        return orth_pos
    
    def orth2frac(self, orth_pos: torch.Tensor) -> torch.Tensor:
        """
        Convert orthogonal coordinates to fractional coordinates

        Args:
            orth_pos: torch.Tensor, [n_points, ..., 3]
        
        Returns:
            fractional coordinates, torch.Tensor
        """
        frac_pos = torch.einsum("n...x,yx->n...y", orth_pos, self.orth2frac_tensor)
        return frac_pos
    
    def exp_sym(self, frac_pos: torch.Tensor | None = None) -> torch.Tensor:
        """
        Apply all symmetry operations to the fractional coordinates

        Args:
            frac_pos, torch.Tensor, [n_points, 3]
                fractional coordinates of model in single ASU. 
                If not given, will use self.atom_pos_frac
        Returns:
            torch.Tensor, [n_points, n_ops, 3]
                fractional coordinates of symmetry operated models        
        """
        if frac_pos is None:
            frac_pos = self.atom_pos_frac
        sym_oped_frac_pos = torch.einsum("oxy,ay->aox", self.R_G_tensor_stack, frac_pos) + self.T_G_tensor_stack
        return sym_oped_frac_pos
    
    def init_mtz(self, mtzdata, N_bins, expcolumns, set_experiment, freeflag, testset_value, dmin):
        """
        set mtz file for HKL list, resolution and experimental related properties
        """

        if type(mtzdata) == str:
            if mtzdata.endswith(".mtz"):
                mtz_reference = rs.read_mtz(mtzdata)
            elif mtzdata.endswith(".cif"):
                mtz_reference = rs.read_cif(mtzdata)
            else:
                raise ValueError("mtzdata should be rs.Dataset object or path str to a mtz/cif file!")
        elif type(mtzdata) == rs.DataSet:
            mtz_reference = mtzdata
        else:
            raise TypeError("mtzdata should be rs.Dataset object or path str to a mtz/cif file!")

        try:
            mtz_reference.dropna(axis=0, subset=expcolumns, inplace=True)
        except:
            raise ValueError(f"{expcolumns} columns not included in the mtz file!")
        
        # Make sure no systematic absent indices are included, and map to asu
        if rs.utils.is_absent(mtz_reference.get_hkls(), mtz_reference.spacegroup).sum() > 0:
            mtz_reference = mtz_reference[~rs.utils.is_absent(mtz_reference.get_hkls(), mtz_reference.spacegroup)]
            print("Found systematic absent indices in the mtz, already dropped!", flush=True)
        mtz_reference.hkl_to_asu(inplace=True, anomalous=self.anomalous)
        
        if self.anomalous:
            # Try to get the wavelength from MTZ file
            try:
                mtz_wavelength = mtz_reference.dataset(0).wavelength
                assert mtz_wavelength > 0.05
                if self.wavelength is not None:
                    assert np.isclose(mtz_wavelength, self.wavelength, atol=0.05)
                else:
                    self.wavelength = mtz_wavelength
            except:
                print(
                    "Can't find wavelength record in the MTZ file, or it doesn't match with other sources"
                )

        if (mtz_reference.cell == self._pdb.cell):
            pass
        else:
            print("Unit cell from mtz file does not match that in PDB file! Using the cell info from MTZ file!")
            self._pdb.set_unitcell(mtz_reference.cell)

        if (mtz_reference.spacegroup.hm == self._pdb.spacegroup.hm):
            pass
        else:
            print("Space group from mtz file does not match that in PDB file! Using the spacegroup from MTZ file!")  # type: ignore
            self._pdb.set_spacegroup(mtz_reference.spacegroup)

        # HKL array from the reference mtz file, [N,3]
        self.HKL_array = mtz_reference.get_hkls()
        self.dHKL = self.unit_cell.calculate_d_array(self.HKL_array).astype(
            "float32"
        )
        if dmin is None:
            self.dmin = self.dHKL.min()
            resol_bool = None
        else:
            resol_bool = self.dHKL >= dmin
            self.dmin = dmin
            self.dHKL = self.dHKL[resol_bool]
            self.HKL_array = self.HKL_array[resol_bool] 
            
        self.Hasu_array = generate_reciprocal_asu(
            self.unit_cell, self.space_group, self.dmin, anomalous=self.anomalous
        )
        assert (
            diff_array(self.HKL_array, self.Hasu_array) == set()
        ), "HKL_array should be equal or subset of the Hasu_array!"
        self.asu2HKL_index = asu2HKL(self.Hasu_array, self.HKL_array)
        # d*^2 array according to the HKL list, [N]
        self.dr2asu_array = self.unit_cell.calculate_1_d2_array(self.Hasu_array)
        self.dr2HKL_array = self.unit_cell.calculate_1_d2_array(self.HKL_array)
        # assign reslution bins
        self.assign_resolution_bins(bins=N_bins)
        if set_experiment:
            self.set_experiment(mtz_reference, expcolumns, freeflag, testset_value, resol_bool)

    def set_experiment(self, exp_mtz, expcolumns=["FP", "SIGFP"], freeflag="FreeR_flag", testset_value=0, resol_bool=None):
        """
        Set experimental data for refinement,
        including Fo, SigF, free_flag, Outlier

        exp_mtz, rs.Dataset, mtzfile read by reciprocalspaceship
        """
        if resol_bool is None:
            resol_bool = np.ones_like(self.dHKL, dtype=bool)
        exp_mtz = exp_mtz[resol_bool].copy()
        try:
            self.Fo = torch.tensor(exp_mtz[expcolumns[0]].to_numpy(), device=self.device).type(
                torch.float32
            )
            self.SigF = torch.tensor(
                exp_mtz[expcolumns[1]].to_numpy(), device=self.device
            ).type(torch.float32)
            nonpositive_Fo = assert_numpy(self.Fo <= 0.0)
            if nonpositive_Fo.sum() > 0:
                print(f"With {nonpositive_Fo.sum()} non-positive experimental magnitudes! Will be treated as outliers.")
        except:
            print(f"MTZ file doesn't contain {expcolumns[0]} or {expcolumns[1]}! Check your data!")

        try:
            self.free_flag = np.where(
                exp_mtz[freeflag].values == testset_value, True, False
            )
        except:
            self.free_flag = np.zeros(len(self.Fo)).astype(bool) 
            print("No Free Flag Column, will treat all elements as working set! Check your data!")

        # label outliers
        # Detecting outliers in non-redundant diffraction data, Read, 1999
        try:
            exp_mtz.label_centrics(inplace=True)
            exp_mtz.compute_multiplicity(inplace=True)
            self.centric = exp_mtz["CENTRIC"].to_numpy()
            self.epsilon = torch.tensor(exp_mtz["EPSILON"].to_numpy(), device=self.device).type(
                torch.float32
            )
            exp_mtz["SIGN"] = 0.0
            for i in range(self.n_bins):
                index_i = self.bins == i
                exp_mtz.loc[index_i, "SIGN"] = np.mean(
                    exp_mtz[index_i][expcolumns[0]].to_numpy() ** 2
                    / exp_mtz[index_i]["EPSILON"].to_numpy()
                )
            exp_mtz["EP"] = exp_mtz[expcolumns[0]] / np.sqrt(exp_mtz["EPSILON"].astype(float) * exp_mtz["SIGN"])
            self._SigN = torch.tensor(exp_mtz["SIGN"].to_numpy(), device=self.device).type(
                torch.float32
            ) 
            self.Eo = torch.tensor(exp_mtz["EP"].to_numpy(), device=self.device).type(
                torch.float32
            )
            self.SigEo = self.SigF / (self._SigN * self.epsilon).sqrt()
            self.Outlier = (
                (exp_mtz["CENTRIC"] & (exp_mtz["EP"] > 4.89))
                | ((~exp_mtz["CENTRIC"]) & (exp_mtz["EP"] > 3.72))
            ).to_numpy(dtype=bool)
        except:
            self.Outlier = np.zeros(len(self.Fo)).astype(bool)
            print("No outlier detection, will use all reflections!")
        self.Outlier = self.Outlier | nonpositive_Fo 
    
    def init_withoutmtz(self, dmin, n_bins):
        if not dmin:
            raise ValueError(
                "high_resolution dmin OR a reference mtz file should be provided!"
            )
        else:
            self.dmin = dmin
            # Generate ASU HKL array and Corresponding d*^2 array
            self.Hasu_array = generate_reciprocal_asu(
                self.unit_cell, self.space_group, self.dmin
            )
            self.dHasu = self.unit_cell.calculate_d_array(self.Hasu_array).astype(
                "float32"
            )
            self.dr2asu_array = self.unit_cell.calculate_1_d2_array(self.Hasu_array)
            self.HKL_array = None
            self.assign_resolution_bins(n_bins)

    def init_atomic_scattering(self):
        """
        mode xray: xray atomic scatterring factor
            A dictionary of atomic structural factor f0_sj of different atom types at different HKL Rupp's Book P280
            f0_sj = [sum_{i=1}^4 {a_ij*exp(-b_ij* d*^2/4)} ] + c_j
        
        mode cryoem: electron atomic scattering factor
            The electron form factors are parametrized as 5 Gaussians, using coefficients from International Tables 
            for Crystallography Volume C, edition 2011, table 4.3.2.2 (pp. 282-283), “Elastic atomic scattering 
            factors of electrons for neutral atoms and s up to 2.0 A–1”. The same parametrization is used in cctbx and in CCP4.
        """
        
        if self.anomalous:
            assert self.mode == "xray", ValueError(
                "Only support anomalous scattering for xray mode!"
            )
            assert self.wavelength is not None, ValueError(
                "If you need anomalous scattering contribution, provide the wavelength info from input, pdb or mtz file!"
            )

        self.full_atomic_sf_asu = {}
        for atom_type in self.unique_atom:
            element = gemmi.Element(atom_type)
            if self.mode == "xray":
                f0 = np.array(
                    [element.it92.calculate_sf(dr2 / 4.0) for dr2 in self.dr2asu_array]
                )
                if self.anomalous:
                    fp, fpp = gemmi.cromer_liberman(
                        z=element.atomic_number, energy=gemmi.hc / self.wavelength
                    )
                    self.full_atomic_sf_asu[atom_type] = f0 + fp + 1j * fpp
                else:
                    self.full_atomic_sf_asu[atom_type] = f0
            elif self.mode == "cryoem":
                f0 = np.array(
                    [element.c4322.calculate_sf(dr2 / 4.0) for dr2 in self.dr2asu_array]
                )
                self.full_atomic_sf_asu[atom_type] = f0

        if self.anomalous:
            self.fullsf_tensor = torch.tensor(
                np.array([self.full_atomic_sf_asu[atom] for atom in self.atom_name]),
                device=self.device,
            ).type(torch.complex64)
        else:
            self.fullsf_tensor = torch.tensor(
                np.array([self.full_atomic_sf_asu[atom] for atom in self.atom_name]),
                device=self.device,
            ).type(torch.float32)


    def assign_resolution_bins(
        self, bins=10, Nmin=100, return_labels=False, format_str=".2f"
    ):
        """Assign reflections in HKL_array to resolution bins with logarithmic algorithm.
        The labels will be stored in self.bins

        Urzhumtsev, A., et al. Acta Crystallographica Section D: Biological Crystallography 65.12 (2009)
        https://doi.org/10.1107/S0907444909039638 

        Parameters
        ----------
        bins : int, optional
            Number of bins, by default 10
        Nmin : int, optional
            Minimum number of reflections for the first two low-resolution ranges, by default 100
        return_labels : bool, optional
            Whether to return a list of labels corresponding to the edges
            of each resolution bin, by default False
        format_str : str, optional
            Format string for constructing bin labels, by default ".2f"

        Returns
        -------
        None or list of labels
        """
        if not self.HKL_array is None:
            assert hasattr(
                self, "dHKL"
            ), "Must have resolution stored in dHKL attribute!"
            d = self.dHKL
        else:
            assert hasattr(
                self, "dHasu"
            ), "Must have resolution stored in dHasu attribute!"
            d = self.dHasu
        assignments, edges = bin_by_logarithmic(d, bins, Nmin)
        self.n_bins = bins
        self.bins = assignments
        self.bin_labels = [
            f"{e1:{format_str}} - {e2:{format_str}}"
            for e1, e2 in zip(edges[:-1], edges[1:])
        ]
        if return_labels:
            return self.bin_labels

    def inspect_data(self, verbose=False, spacing=4.5):
        """
        Do an inspection of data, for hints about
        1. solvent percentage for mask calculation
        2. suitable grid size
        """
        # solvent percentage
        vdw_rad = vdw_rad_tensor(self.atom_name, device=self.device)
        uc_grid_orth_tensor = unitcell_grid_center(
            self.unit_cell, spacing=spacing, return_tensor=True, device=self.device
        )
        occupancy, _ = packingscore_voxelgrid_torch(
            self.atom_pos_orth,
            self.unit_cell,
            self.space_group,
            vdw_rad,
            uc_grid_orth_tensor,
        )
        self.solventpct = 1 - occupancy
        # grid size
        mtz = gemmi.Mtz(with_base=True)
        mtz.cell = self.unit_cell
        mtz.spacegroup = self.space_group
        if not self.HKL_array is None:
            mtz.set_data(self.HKL_array)
        else:
            mtz.set_data(self.Hasu_array)
        self.gridsize = mtz.get_size_for_hkl(sample_rate=3.0)
        if verbose:
            print("Solvent Percentage:", self.solventpct)
            print("Grid size:", self.gridsize)
        self.inspected = True

    def calc_fprotein(
        self,
        atoms_position_tensor=None,
        atoms_biso_tensor=None,
        atoms_aniso_uw_tensor=None,
        atoms_occ_tensor=None,
        Return=False,
    ):
        """
        Calculate the structural factor from a single atomic model, without solvent masking

        Parameters
        ----------
        atoms_positions_tensor: 2D [N_atoms, 3] tensor or default None
            Positions of atoms in the model, in unit of angstrom; If not given, the model stored in attribute `atom_pos_orth` will be used

        atoms_biso_tensor: 1D [N_atoms,] tensor or default None
            Isotropic B factors of each atoms in the model; If not given, the info stored in attribute `atoms_b_iso` will be used

        atoms_aniso_uw_tensor: 3D [N_atoms, 3, 3] tensor or default None
            Anisotropic B factors of each atoms in the model, in matrix form; If not given, the info stored in attribute `atoms_aniso_uw` will be used

        atoms_occ_tensor: 1D [N_atoms,] tensor or default None
            Occupancy of each atoms in the model; If not given, the info stored in attribute `atom_occ` will be used

        NO_Bfactor: Boolean, default False
            If True, the calculation will not use Bfactor parameterization; Useful when we are parameterizing the ensemble with a true distribution

        Return: Boolean, default False
            If True, it will return the Fprotein as the function output; Or It will just be saved in the `Fprotein_asu` and `Fprotein_HKL` attributes

        Returns
        -------
        None (Return=False) or Fprotein (Return=True)
        """
        # Read and tensor-fy necessary inforamtion
        if not atoms_position_tensor is None:
            self.atom_pos_orth = atoms_position_tensor

        if not atoms_aniso_uw_tensor is None:
            self.atom_aniso_uw = atoms_aniso_uw_tensor

        if not atoms_biso_tensor is None:
            self.atom_b_iso = atoms_biso_tensor

        if not atoms_occ_tensor is None:
            self.atom_occ = atoms_occ_tensor

        self.Fprotein_asu = F_protein(
            self.Hasu_array,
            self.dr2asu_array,
            self.fullsf_tensor,
            self.R_G_tensor_stack,
            self.T_G_tensor_stack,
            self.orth2frac_tensor,
            self.atom_pos_frac,
            self.atom_b_iso,
            self.atom_aniso_uw,
            self.atom_occ,
        )
        if not self.HKL_array is None:
            self.Fprotein_HKL = self.Fprotein_asu[self.asu2HKL_index]
            self.Fmask_HKL = torch.zeros_like(self.Fprotein_HKL)
            if Return:
                return self.Fprotein_HKL
        else:
            self.Fmask_asu = torch.zeros_like(self.Fprotein_asu)
            if Return:
                return self.Fprotein_asu

    def calc_fsolvent(
        self,
        solventpct=None,
        gridsize=None,
        dmin_mask=5.0,
        dmin_nonzero=3.0,
        exponent=10.0,
        Return=False,
    ):
        """
        Calculate the structure factor of solvent mask in a differentiable way

        Parameters
        ----------
        solventpct: 0 - 1 Float, default None
            An approximate value of volume percentage of solvent in the unitcell.
            run `inspect_data` before to use a suggested value

        gridsize: [Int, Int, Int], default None
            The size of grid to construct mask map.
            run `inspect_data` before to use a suggected value

        dmin_mask: np.float32, Default 6 angstroms.
            Minimum resolution cutoff, in angstroms, for creating the solvent mask

        Return: Boolean, default False
            If True, it will return the Fmask as the function output; Or It will just be saved in the `Fmask_asu` and `Fmask_HKL` attributes
        """

        if solventpct is None:
            assert (
                self.inspected
            ), "Run inspect_data first or give a valid solvent percentage!"
            solventpct = self.solventpct

        if gridsize is None:
            assert self.inspected, "Run inspect_data first or give a valid grid size!"
            gridsize = self.gridsize

        # Shape [N_HKL_p1, 3], [N_HKL_p1,]
        Hp1_array, Fp1_tensor = expand_to_p1(
            self.space_group,
            self.Hasu_array,
            self.Fprotein_asu,
            dmin_mask=dmin_mask,
            unitcell=self.unit_cell,
            anomalous=self.anomalous,
        )
        rs_grid = reciprocal_grid(Hp1_array, Fp1_tensor, gridsize)
        self.real_grid_mask = rsgrid2realmask(
            rs_grid, solvent_percent=solventpct, exponent=exponent, 
        )  # type: ignore
        if not self.HKL_array is None:
            self.Fmask_HKL = realmask2Fmask(self.real_grid_mask, self.HKL_array)
            zero_hkl_bool = torch.tensor(self.dHKL <= dmin_nonzero, device=self.device)
            self.Fmask_HKL[zero_hkl_bool] = torch.tensor(
                0.0, device=self.device, dtype=torch.complex64
            )
            if Return:
                return self.Fmask_HKL
        else:
            self.Fmask_asu = realmask2Fmask(self.real_grid_mask, self.Hasu_array)
            zero_hkl_bool = torch.tensor(self.dHasu <= dmin_nonzero, device=self.device)
            self.Fmask_asu[zero_hkl_bool] = torch.tensor(
                0.0, device=self.device, dtype=torch.complex64
            )
            if Return:
                return self.Fmask_asu

    def _init_kmask_kiso(self, requires_grad=True):
        """
        Use the root finding approach discussed to initialize kmask and kiso per resolution bin
        Afonine, P. V., et al. Acta Crystallographica Section D: Biological Crystallography 69.4 (2013): 625-634.

        Note: Only work when you have mtz data, self.Fo
        """

        kmasks = []
        kisos = []
        Fmask_HKL = self.Fmask_HKL.detach().clone()
        Fprotein_HKL = self.Fprotein_HKL.detach().clone()
        ws = torch.abs(Fmask_HKL).pow(2)
        vs = 0.5 * torch.real(
            Fprotein_HKL.conj() * Fmask_HKL
            + Fprotein_HKL * Fmask_HKL.conj()
        )
        us = torch.abs(Fprotein_HKL).pow(2)
        Is = self.Fo.pow(2)

        for bin_i in np.sort(np.unique(self.bins)):
            index_i = (~self.free_flag) & (self.bins == bin_i) & (~self.Outlier)

            C2 = torch.sum(ws[index_i] * Is[index_i])
            B2 = 2.0 * torch.sum(vs[index_i] * Is[index_i])
            A2 = torch.sum(us[index_i] * Is[index_i])
            Y2 = torch.sum(Is[index_i].pow(2))
            # They made this wrong in their original paper
            D3 = torch.sum(ws[index_i].pow(2))
            C3 = 3.0 * torch.sum(ws[index_i] * vs[index_i])
            B3 = torch.sum(2 * vs[index_i].pow(2) + us[index_i] * ws[index_i])
            A3 = torch.sum(us[index_i] * vs[index_i])
            Y3 = torch.sum(Is[index_i] * vs[index_i])

            a = assert_numpy((C3 * Y2 - C2 * B2 - C2 * Y3) / (D3 * Y2 - C2**2))
            b = assert_numpy((B3 * Y2 - C2 * A2 - Y3 * B2) / (D3 * Y2 - C2**2))
            c = assert_numpy((A3 * Y2 - Y3 * A2) / (D3 * Y2 - C2**2))

            try:
                roots = np.roots([1.0, a, b, c])
            except:
                # In case no nonzero Fmask observations for this resolution bin
                kmask = torch.tensor(1e-4).to(C2)
                K_temp = (kmask.pow(2) * C2 + kmask * B2 + A2) / Y2
                kiso = torch.sqrt(K_temp).reciprocal()
                kmasks.append(kmask.requires_grad_())
                kisos.append(kiso.requires_grad_())
                continue

            kmask_candidates = torch.zeros(3).to(C2)
            LSpp_candidates = torch.zeros(3).to(C2)
            K_candidates = torch.zeros(3).to(C2)
            LS_candidates = torch.zeros(3).to(C2)
            for i, root in enumerate(roots):
                if np.iscomplex(root) or (root < 0):
                    kmask_temp = torch.tensor(0.0001).to(C2)
                else:
                    kmask_temp = torch.tensor(np.real_if_close(root, tol=1000)).to(C2)
                K_temp = (kmask_temp.pow(2) * C2 + kmask_temp * B2 + A2) / Y2
                if K_temp < 0.0:
                    kmask_temp = torch.tensor(1.0).to(C2)
                LS_temp = torch.sum(
                    torch.abs(
                        self.Fprotein_HKL[index_i]
                        + kmask_temp * self.Fmask_HKL[index_i]
                    ).pow(2)
                    - K_temp * Is[index_i]
                ).pow(2)
                LSpp_temp = (
                    3 * kmask_temp.pow(2) * D3 + 2 * kmask_temp * C3 + B3 - C2 * K_temp
                )
                kmask_candidates[i] = kmask_temp
                # Note, K should be positive too
                K_candidates[i] = K_temp
                LSpp_candidates[i] = LSpp_temp
                LS_candidates[i] = LS_temp

            legitimacy = (
                (roots > 0) & np.isreal(roots) & (assert_numpy(LSpp_candidates) > 0)
            )
            if (np.sum(legitimacy) > 0) and (np.sum(legitimacy) < 3):
                kmask_candidates = kmask_candidates[legitimacy]
                LS_candidates = LS_candidates[legitimacy]
                K_candidates = K_candidates[legitimacy]

            min_index = torch.argmin(LS_candidates)
            kmask = kmask_candidates[min_index]
            # K = k_total^(-2)
            kiso = torch.sqrt(K_candidates[min_index]).reciprocal()

            kmasks.append(kmask.requires_grad_(requires_grad))
            kisos.append(kiso.requires_grad_(requires_grad))
        return kmasks, kisos

    def _init_uaniso(self, requires_grad=True):
        """
        Use the analytical solutuon discussed to initialize Uaniso per resolution bin
        Afonine, P. V., et al. Acta Crystallographica Section D: Biological Crystallography 69.4 (2013): 625-634.

        Note: Only work when you have mtz data, self.Fo
        """
        uanisos = []
        for bin_i in np.sort(np.unique(self.bins)):
            index_i = (self.bins == bin_i) & (~self.free_flag) & (~self.Outlier)
            s = self.HKL_array[index_i]
            V = np.concatenate([s**2, 2 * s[:, [0, 2, 1]] * s[:, [1, 0, 2]]], axis=-1)
            Z = assert_numpy(
                torch.log(
                    self.Fo[index_i]
                    / (
                        self.kisos[bin_i]
                        * torch.abs(
                            self.Fprotein_HKL[index_i]
                            + self.kmasks[bin_i] * self.Fmask_HKL[index_i]
                        )
                    )
                )
                / (2.0 * np.pi**2)
            )
            M = V.T @ V  # M = np.einsum("ki,kj->ij", V, V)
            b = -np.sum(Z * V.T, axis=-1)
            U = np.linalg.inv(M) @ b
            uanisos.append(torch.tensor(U).to(self.Fo).requires_grad_(requires_grad))
        return uanisos

    def _set_scales(
        self,
        requires_grad,
        kiso=1.0,
        kmask=0.35,
        uaniso=[0.01, 0.01, 0.01, 1e-4, 1e-4, 1e-4],
    ):
        """Only used for case you don't have data"""
        self.kmasks = [
            torch.tensor(kmask).to(self.atom_pos_frac).requires_grad_(requires_grad)
            for i in range(self.n_bins)
        ]
        self.kisos = [
            torch.tensor(kiso).to(self.atom_pos_frac).requires_grad_(requires_grad)
            for i in range(self.n_bins)
        ]
        self.uanisos = [
            torch.tensor(uaniso).to(self.atom_pos_frac).requires_grad_(requires_grad)
            for i in range(self.n_bins)
        ]

    def init_scales(self, requires_grad=True):
        if hasattr(self, "Fo"):
            self.kmasks, self.kisos = self._init_kmask_kiso(requires_grad=requires_grad)
            self.uanisos = self._init_uaniso(requires_grad=requires_grad)
            Fmodel = self.calc_ftotal()
            self.r_work, self.r_free = self.get_rfactors(ftotal=Fmodel)
        else:
            self._set_scales(requires_grad)

    def _get_scales_lbfgs_LS(
        self, n_steps=3, lr=0.1, verbose=True, initialize=True, return_loss=False
    ):
        """
        Use LBFGS to optimize scales with least square error
        """
        assert hasattr(self, "Fo"), "No experimental data Fo!"

        if initialize:
            self.init_scales(requires_grad=True)
        else:
            self.unfreeze_scales()

        def closure():
            Fmodel = self.calc_ftotal(scale_mode=True)
            Fmodel_mag = torch.abs(Fmodel)
            # LS loss
            loss = torch.sum(
                (
                    self.Fo[(~self.free_flag) & (~self.Outlier)]
                    - Fmodel_mag[(~self.free_flag) & (~self.Outlier)]
                )
                ** 2
            )
            self.lbfgs.zero_grad()
            loss.backward()
            return loss

        params = self.kmasks + self.kisos + self.uanisos
        self.lbfgs = torch.optim.LBFGS(params, lr=lr)
        loss_track = []
        for _ in range(n_steps):
            start_time = time.time()
            loss = self.lbfgs.step(closure)
            Fmodel = self.calc_ftotal()
            r_work, r_free = self.get_rfactors(ftotal=Fmodel)
            loss_track.append(
                [assert_numpy(loss), assert_numpy(r_work), assert_numpy(r_free)]
            )
            str_ = f"Time: {time.time()-start_time:.3f}"
            if verbose:
                print(
                    f"Scale, {loss_track[-1][0]:.3f}, {loss_track[-1][1]:.3f}, {loss_track[-1][2]:.3f}",
                    str_,
                    flush=True,
                )
        self.r_work, self.r_free = r_work, r_free
        if return_loss:
            return loss_track

    def _get_scales_lbfgs_r(
        self, n_steps=5, lr=0.1, verbose=True, initialize=False, return_loss=False
    ):
        """
        Use LBFGS to optimize scales directly with r factor error
        """
        assert hasattr(self, "Fo"), "No experimental data Fo!"

        if initialize:
            self.init_scales(requires_grad=True)
        else:
            self.unfreeze_scales()

        def closure():
            Fmodel = self.calc_ftotal(scale_mode=True)
            Fmodel_mag = torch.abs(Fmodel)
            # R factor
            loss = torch.sum(
                torch.abs(
                    self.Fo[(~self.free_flag) & (~self.Outlier)]
                    - Fmodel_mag[(~self.free_flag) & (~self.Outlier)]
                )
            ) / torch.sum(self.Fo[(~self.free_flag) & (~self.Outlier)])
            self.lbfgs.zero_grad()
            loss.backward()
            return loss

        params = self.kmasks + self.kisos + self.uanisos
        self.lbfgs = torch.optim.LBFGS(params, lr=lr)
        loss_track = []
        for _ in range(n_steps):
            start_time = time.time()
            loss = self.lbfgs.step(closure)
            Fmodel = self.calc_ftotal()
            r_work, r_free = self.get_rfactors(ftotal=Fmodel)
            loss_track.append(
                [assert_numpy(loss), assert_numpy(r_work), assert_numpy(r_free)]
            )
            str_ = f"Time: {time.time()-start_time:.3f}"
            if verbose:
                print(
                    f"Scale, {loss_track[-1][0]:.3f}, {loss_track[-1][1]:.3f}, {loss_track[-1][2]:.3f}",
                    str_,
                    flush=True,
                )
        self.r_work, self.r_free = r_work, r_free
        if return_loss:
            return loss_track

    def get_scales_lbfgs(
        self,
        ls_steps=3,
        r_steps=3,
        ls_lr=0.1,
        r_lr=0.1,
        initialize=True,
        verbose=False,
    ):
        self._get_scales_lbfgs_LS(ls_steps, ls_lr, verbose, initialize)
        self._get_scales_lbfgs_r(r_steps, r_lr, verbose, initialize=False)
        self.freeze_scales()

    def freeze_scales(self):
        """
        Do not require grad on scales
        """
        self.kmasks = [kmask.requires_grad_(False) for kmask in self.kmasks]
        self.kisos = [kiso.requires_grad_(False) for kiso in self.kisos]
        self.uanisos = [uaniso.requires_grad_(False) for uaniso in self.uanisos]
    
    def unfreeze_scales(self):
        """
        Do not require grad on scales
        """
        self.kmasks = [kmask.requires_grad_(True) for kmask in self.kmasks]
        self.kisos = [kiso.requires_grad_(True) for kiso in self.kisos]
        self.uanisos = [uaniso.requires_grad_(True) for uaniso in self.uanisos]

    def get_scales_adam(
        self, 
        lr=0.01, 
        n_steps=100, 
        sub_ratio=0.7, 
        initialize=True, 
        verbose=False
    ):
        def adam_opt_i(
            bin_i, index_i, sub_ratio=0.3, lr=0.001, n_steps=100, verbose=False
        ):
            def adam_stepopt(sub_boolean_mask):
                Fmodel_i = self._calc_ftotal_bini(
                    bin_i, index_i, self.HKL_array, self.Fprotein_HKL, self.Fmask_HKL, scale_mode=True
                )
                Fmodel_mag = torch.abs(Fmodel_i)
                # LS loss with subsampling
                fo_i_sub = fo_i[sub_boolean_mask]
                Fmodel_i_sub = Fmodel_mag[sub_boolean_mask]
                free_flagi_sub = free_flagi[sub_boolean_mask]
                loss = torch.sum(
                    (fo_i_sub[~free_flagi_sub] - Fmodel_i_sub[~free_flagi_sub]) ** 2
                )
                # loss = torch.sum(torch.abs(fo_i_sub[~free_flagi_sub] - Fmodel_i_sub[~free_flagi_sub]))/torch.sum(Fmodel_i_sub[~free_flagi_sub])
                r_work, r_free = r_factor(fo_i, Fmodel_mag, free_flagi)
                adam.zero_grad()
                loss.backward()
                adam.step()
                return loss, r_work, r_free

            params = [self.kmasks[bin_i], self.kisos[bin_i], self.uanisos[bin_i]]
            adam = torch.optim.Adam(params, lr=lr)
            for _ in range(n_steps):
                start_time = time.time()
                sub_boolean_mask = (
                    np.random.rand(
                        np.sum(index_i),
                    )
                    < sub_ratio
                )
                temp = adam_stepopt(sub_boolean_mask)
                time_this_round = round(time.time() - start_time, 3)
                str_ = "Time: " + str(time_this_round)
                if verbose:
                    print("Scale", *[assert_numpy(i) for i in temp], str_, flush=True)

        if initialize:
            self.init_scales(requires_grad=True)
        else:
            self.unfreeze_scales()

        for bin_i in range(self.n_bins):
            index_i = (self.bins == bin_i) & (~self.Outlier)
            fo_i = self.Fo[index_i]
            free_flagi = self.free_flag[index_i]
            adam_opt_i(
                bin_i,
                index_i,
                lr=lr,
                n_steps=n_steps,
                sub_ratio=sub_ratio,
                verbose=verbose,
            )
        self.freeze_scales()
        Fmodel = self.calc_ftotal()
        self.r_work, self.r_free = self.get_rfactors(ftotal=Fmodel)

    def _calc_ftotal_bini(self, bin_i, index_i, HKL_array, Fprotein, Fmask, scale_mode=False):
        """
        calculate ftotal for bin i
        """
        if scale_mode:
            Fmask = Fmask.detach()
            Fprotein = Fprotein.detach()
        scaled_fmask_i = Fmask[index_i] * self.kmasks[bin_i]
        fmodel_i = (
            self.kisos[bin_i]
            * aniso_scaling(
                self.uanisos[bin_i],
                self.reciprocal_cell_paras,
                HKL_array[index_i],
            )
            * (Fprotein[index_i] + scaled_fmask_i)
        )
        return fmodel_i

    def calc_ftotal(self, bins=None, Return=True, scale_mode=False):
        """
        Calculate Ftotal = kiso * exp(-2*pi^2*s^T*Uaniso*s) * (Fprotein + kmask * Fmask)

        kiso, uaniso and kmask are stored for each resolution bin

        Parameters
        ----------
        bins: None or List[int], default None
            Specify which resolution bins to calculate the ftotal, if None, calculate for all

        Return: Boolean, default True
            Whether to return the results

        Returns
        -------
        torch.tensor, complex
        """
        if bins is None:
            bins = range(self.n_bins)

        if not self.HKL_array is None:
            ftotal_hkl = torch.zeros_like(self.Fprotein_HKL)
            for bin_i in bins:
                index_i = self.bins == bin_i
                ftotal_hkl[index_i] = self._calc_ftotal_bini(
                    bin_i, index_i, self.HKL_array, self.Fprotein_HKL, self.Fmask_HKL, scale_mode=scale_mode
                )
            self.Ftotal_HKL = ftotal_hkl
            if Return:
                return ftotal_hkl
        else:
            ftotal_asu = torch.zeros_like(self.Fprotein_asu)
            for bin_i in bins:
                index_i = self.bins == bin_i
                ftotal_asu[index_i] = self._calc_ftotal_bini(
                    bin_i, index_i, self.Hasu_array, self.Fprotein_asu, self.Fmask_asu, scale_mode=scale_mode
                )
            self.Ftotal_asu = ftotal_asu
            if Return:
                return ftotal_asu
    
    def calc_Ec(self, Ftotal_HKL=None):
        """
        compute normalized magnitudes from Fc
        
        Args:
            Ftotal_HKL (torch.Tensor) : [N_HKL] or [N_batch, N_HKL]
        
        Returns:
            Ec_HKL : [N_HKL] or [N_batch, N_HKL]
        """ 
        if Ftotal_HKL is None:
            Ftotal_HKL = self.Ftotal_HKL

        SigmaP = torch.zeros_like(Ftotal_HKL, dtype=torch.float32)
        for i in range(self.n_bins):
            index_i = self.bins == i
            SigmaP[..., index_i] = torch.mean(
                Ftotal_HKL[..., index_i].abs().square()
                / self.epsilon[index_i], dim=-1, keepdims=True
            )
        Ec_HKL = Ftotal_HKL / torch.sqrt(self.epsilon * SigmaP)
        return Ec_HKL
    
    def get_sigmaA(self, Ec_HKL, subset="working"):
        """
        Compute sigmaA using corr(Eo, Ec)

        Args:
            Ec_HKL (torch.tensor) : [N_HKL] or [N_batch, N_HKL]
        
        Returns:
            sigmaAs [N_bins] or [N_batch, N_bins]
        """
        if subset == "working":
            subset_boolean = (~self.free_flag) & (~self.Outlier)
        else:
            subset_boolean = (~self.Outlier)
        Ecalc = Ec_HKL.abs().detach().clone()
        sigmaAs = torch.zeros(*Ecalc.shape[:-1], self.n_bins, dtype=torch.float32, device=self.device)
        for i in range(self.n_bins):
            index_i = self.bins[subset_boolean] == i
            Eobs_i = self.Eo[subset_boolean][index_i].square() # [N_subset]
            Ecalc_i = Ecalc[..., subset_boolean][..., index_i].square() # [N_batch, N_subset] or [N_subset]
            # Compute correlation coefficient
            Eoi_centered = Eobs_i - Eobs_i.mean()
            Eci_centered = Ecalc_i - torch.mean(Ecalc_i, dim=-1, keepdims=True)
            Covi = (Eci_centered @ Eoi_centered) / (Eoi_centered.shape[0] -1)
            Eoi_std = torch.std(Eoi_centered, correction=0)
            Eci_std = torch.std(Eci_centered, dim=-1, correction=0)
            sigmaAs[..., i] = (Covi / (Eoi_std * Eci_std)).clamp(min=0.001, max=0.999).sqrt()
        return sigmaAs

    def summarize(self, Ftotal_HKL=None):
        """
        Print model quality log like phenix.model_vs_data
        """
        if Ftotal_HKL is None:
            ftotal = self.calc_ftotal(Return=True)
        else:
            ftotal = Ftotal_HKL
        _, counts = np.unique(self.bins, return_counts=True)
        print(
            f"{'Resolution':15} {'N_work':>7} {'N_free':>7} {'<Fobs>':>7} {'<|Fmodel|>':>9} {'R_work':>7} {'R_free':>7} {'k_mask':>7} {'k_iso':>7}"
        )
        for i in range(self.n_bins):
            index_i = (self.bins == i) & (~self.Outlier)
            r_worki, r_freei = r_factor(
                self.Fo[index_i], torch.abs(ftotal[index_i]), self.free_flag[index_i]
            )
            N_work = counts[i] - np.sum(self.free_flag[index_i])
            N_free = np.sum(self.free_flag[index_i])
            print(
                f"{self.bin_labels[i]:<15} {N_work:7d} {N_free:7d} {assert_numpy(torch.mean(self.Fo[index_i])):7.1f} {assert_numpy(torch.mean(torch.abs(ftotal[index_i]))):9.1f} {assert_numpy(r_worki):7.3f} {assert_numpy(r_freei):7.3f} {assert_numpy(self.kmasks[i]):7.3f} {assert_numpy(self.kisos[i]):7.3f}"
            )
        self.r_work, self.r_free = self.get_rfactors(ftotal=ftotal)
        print(f"r_work: {assert_numpy(self.r_work):<7.3f}")
        print(f"r_free: {assert_numpy(self.r_free):<7.3f}")
        print(f"Number of outliers: {np.sum(self.Outlier):<7d}")
    
    def get_rfactors(self, ftotal=None, mask=None):
        if ftotal is None:
            ftotal = self.Ftotal_HKL.detach().clone()
        if mask is None:
            mask = np.ones(len(self.Fo)).astype(bool)
        bool_array = (~self.Outlier) & mask
        r_work, r_free = r_factor(
            self.Fo[bool_array],
            torch.abs(ftotal)[bool_array],
            self.free_flag[bool_array],
        )
        return r_work, r_free

    def calc_fprotein_batch(self, atoms_position_batch, Return=False, PARTITION=20):
        """
        Calculate the Fprotein with batched models. Most parameters are similar to `Calc_Fprotein`

        atoms_positions_batch: torch.float32 tensor, [N_batch, N_atoms, 3]

        PARTITION: Int, default 20
            To reduce the memory cost during the computation, we divide the batch into several partitions and loops through them.
            Larger PARTITION will require larger GPU memory. Default 20 will require around 21GB, if N_atoms~2400 and N_HKLs~24000. 
            Memory scales linearly with PARITION, N_atoms, and N_HKLs.
            But larger PARTITION will give a smaller wall time, so this is a trade-off.
        """
        # Read and tensor-fy necessary information
        atom_pos_frac_batch = torch.tensordot(
            atoms_position_batch, self.orth2frac_tensor.T, 1
        )  # [N_batch, N_atoms, N_dim=3]

        self.Fprotein_asu_batch = F_protein_batch(
            self.Hasu_array,
            self.dr2asu_array,
            self.fullsf_tensor,
            self.R_G_tensor_stack,
            self.T_G_tensor_stack,
            self.orth2frac_tensor,
            atom_pos_frac_batch,
            self.atom_b_iso,
            self.atom_aniso_uw,
            self.atom_occ,
            PARTITION=PARTITION,
        )  # [N_batch, N_Hasus]

        if not self.HKL_array is None:
            # type: ignore
            self.Fprotein_HKL_batch = self.Fprotein_asu_batch[:, self.asu2HKL_index]
            self.Fmask_HKL_batch = torch.zeros_like(self.Fprotein_HKL_batch)
            if Return:
                return self.Fprotein_HKL_batch
        else:
            self.Fmask_asu_batch = torch.zeros_like(self.Fprotein_asu_batch)
            if Return:
                return self.Fprotein_asu_batch

    def calc_fsolvent_batch(
        self,
        solventpct=None,
        gridsize=None,
        dmin_mask=5,
        dmin_nonzero=3.0,
        exponent=10.0,
        Return=False,
        PARTITION=10,
    ):
        """
        Should run after Calc_Fprotein_batch, calculate the solvent mask structure factors in batched manner
        most parameters are similar to `Calc_Fmask`

        PARTITION: Int, default 10
            To reduce the memory cost during the computation, we divide the batch into several partitions and loops through them.
            Larger PARTITION will require larger GPU memory. Default 10 will take around 2GB, if gridsize=[160,160,160].
            Larger PARTITION does not mean smaller wall time, the optimal size is around 10
        """

        if solventpct is None:
            assert (
                self.inspected
            ), "Run inspect_data first or give a valid solvent percentage!"
            solventpct = self.solventpct

        if gridsize is None:
            assert self.inspected, "Run inspect_data first or give a valid grid size!"
            gridsize = self.gridsize

        Hp1_array, Fp1_tensor_batch = expand_to_p1(
            self.space_group,
            self.Hasu_array,
            self.Fprotein_asu_batch,
            dmin_mask=dmin_mask,
            Batch=True,
            unitcell=self.unit_cell,
        )

        batchsize = self.Fprotein_asu_batch.shape[0]  # type: ignore
        N_partition = batchsize // PARTITION + 1
        Fmask_batch = 0.0

        if not self.HKL_array is None:
            HKL_array = self.HKL_array
        else:
            HKL_array = self.Hasu_array

        for j in range(N_partition):
            if j * PARTITION >= batchsize:
                continue
            start = j * PARTITION
            end = min((j + 1) * PARTITION, batchsize)
            # Shape [N_batch, *gridsize]
            rs_grid = reciprocal_grid(
                Hp1_array, Fp1_tensor_batch[start:end], gridsize, end - start
            )
            real_grid_mask = rsgrid2realmask(
                rs_grid, solvent_percent=solventpct, exponent=exponent, Batch=True
            )  # type: ignore
            Fmask_batch_j = realmask2Fmask(real_grid_mask, HKL_array, end - start)
            if j == 0:
                Fmask_batch = Fmask_batch_j
            else:
                # Shape [N_batches, N_HKLs]
                Fmask_batch = torch.concat(
                    (Fmask_batch, Fmask_batch_j), dim=0
                )  # type: ignore

        if not self.HKL_array is None:
            zero_hkl_bool = torch.tensor(self.dHKL <= dmin_nonzero, device=self.device)
            Fmask_batch[:, zero_hkl_bool] = torch.tensor(
                0.0, device=self.device, dtype=torch.complex64
            )  # type: ignore
            self.Fmask_HKL_batch = Fmask_batch
            if Return:
                return self.Fmask_HKL_batch
        else:
            zero_hkl_bool = torch.tensor(self.dHasu <= dmin_nonzero, device=self.device)
            Fmask_batch[:, zero_hkl_bool] = torch.tensor(
                0.0, device=self.device, dtype=torch.complex64
            )  # type: ignore
            self.Fmask_asu_batch = Fmask_batch
            if Return:
                return self.Fmask_asu_batch

    def _calc_ftotal_batch_bini(self, bin_i, index_i, HKL_array, Fprotein, Fmask):
        """
        calculate ftotal for bin i
        """
        scaled_fmask_i = Fmask[:, index_i] * self.kmasks[bin_i]
        fmodel_i = (
            self.kisos[bin_i]
            * aniso_scaling(
                self.uanisos[bin_i],
                self.reciprocal_cell_paras,
                HKL_array[index_i],
            )
            * (Fprotein[:, index_i] + scaled_fmask_i)
        )
        return fmodel_i

    def calc_ftotal_batch(self, bins=None, Return=True):
        """
        Calculate Ftotal = kiso * exp(-2*pi^2*s^T*Uaniso*s) * (Fprotein + kmask * Fmask)

        Parameters
        ----------
        bins: None or List[int], default None
            Specify which resolution bins to calculate the ftotal, if None, calculate for all

        Return: Boolean, default True
            Whether to return the results

        Returns
        -------
        torch.tensor, complex

        """
        if bins is None:
            bins = range(self.n_bins)
        if not self.HKL_array is None:
            ftotal_hkl_batch = torch.zeros_like(self.Fprotein_HKL_batch)
            for bin_i in bins:
                index_i = self.bins == bin_i
                ftotal_hkl_batch[:, index_i] = self._calc_ftotal_batch_bini(
                    bin_i,
                    index_i,
                    self.HKL_array,
                    self.Fprotein_HKL_batch,
                    self.Fmask_HKL_batch,
                )
            self.Ftotal_HKL_batch = ftotal_hkl_batch
            if Return:
                return ftotal_hkl_batch
        else:
            ftotal_asu_batch = torch.zeros_like(self.Fprotein_asu_batch)
            for bin_i in bins:
                index_i = self.bins == bin_i
                ftotal_asu_batch[index_i] = self._calc_ftotal_batch_bini(
                    bin_i, index_i, self.Hasu_array, self.Fprotein_asu_batch, self.Fmask_asu_batch
                )
            self.Ftotal_asu_batch = ftotal_asu_batch
            if Return:
                return ftotal_asu_batch

    def prepare_dataset(self, HKL_attr: str, F_attr: str):
        """
        create a rs.Dataset with HKL list and complex structure factor
        """
        F_out = getattr(self, F_attr)
        HKL_out = getattr(self, HKL_attr)
        assert len(F_out) == len(
            HKL_out
        ), "HKL and structural factor should have same length!"
        F_out_mag = torch.abs(F_out)
        PI_on_180 = 0.017453292519943295
        F_out_phase = torch.angle(F_out) / PI_on_180
        dataset = rs.DataSet(
            spacegroup=self.space_group, cell=self.unit_cell
        )  # type: ignore
        dataset["H"] = HKL_out[:, 0]
        dataset["K"] = HKL_out[:, 1]
        dataset["L"] = HKL_out[:, 2]
        dataset["FMODEL"] = assert_numpy(F_out_mag)
        dataset["PHIFMODEL"] = assert_numpy(F_out_phase)
        dataset.set_index(["H", "K", "L"], inplace=True)
        dataset.infer_mtz_dtypes(inplace=True)
        return dataset
    
    def savePDB(self, outname: str):
        """
        Write out a PDB with current model coordinates
        """
        
        self._pdb.set_positions(assert_numpy(self.atom_pos_orth))
        self._pdb.set_biso(assert_numpy(self.atom_b_iso))
        self._pdb.set_baniso(assert_numpy(self.atom_aniso_uw))
        self._pdb.set_occ(assert_numpy(self.atom_occ))
        
        from datetime import date
        from . import __version__, __author__
        pdb_header = [
            f'REMARK   1 CREATED WITH SFcalculator {__version__}, {date.today()}',
            f'REMARK   1 AUTHOR {__author__}',
        ]
        self._pdb.pdb_header = pdb_header
        self._pdb.savePDB(outname)

def F_protein(
    HKL_array,
    dr2_array,
    fullsf_tensor,
    R_G_tensor_stack,
    T_G_tensor_stack,
    orth2frac_tensor,
    atom_pos_frac,
    atom_b_iso,
    atom_aniso_uw,
    atom_occ,
):
    """
    Calculate Protein Structural Factor from an atomic model

    atom_pos_frac: 2D tensor, [N_atom, N_dim=3]
    """
    # F_calc = sum_Gsum_j{ [f0_sj*DWF*exp(2*pi*i*(h,k,l)*(R_G*(x1,x2,x3)+T_G))]} fractional postion, Rupp's Book P279
    # G is symmetry operations of the spacegroup and j is the atoms
    # DWF is the Debye-Waller Factor, has isotropic and anisotropic version, based on the PDB file input, Rupp's Book P641
    HKL_tensor = torch.tensor(HKL_array, dtype=torch.float32, device=fullsf_tensor.device)

    oc_sf = fullsf_tensor * atom_occ[..., None]  # [N_atom, N_HKLs]
    # DWF calculator
    dwf_iso = DWF_iso(atom_b_iso, dr2_array)  # [N_atoms, N_HKLs]
    mask_vec = torch.all(torch.all(atom_aniso_uw == 0.0, dim=-1), dim=-1)
    # Vectorized phase calculation
    # sym_oped_pos_frac = (
    #     torch.permute(torch.tensordot(R_G_tensor_stack, atom_pos_frac.T, 1), [2, 0, 1])
    #     + T_G_tensor_stack
    # )
    # Shape [N_atom, N_op, N_dim=3]
    sym_oped_pos_frac = (
        torch.einsum("oxy,ay->aox", R_G_tensor_stack, atom_pos_frac) + T_G_tensor_stack
    )
    sym_oped_hkl = torch.einsum("rx,oxy->roy", HKL_tensor, R_G_tensor_stack)
    exp_phase = 0.0
    # Loop through symmetry operations instead of fully vectorization, to reduce the memory cost
    for i in range(sym_oped_pos_frac.size(dim=1)):
        phase_G = (
            2
            * np.pi
            * torch.einsum("ax,rx->ar", sym_oped_pos_frac[:, i, :], HKL_tensor)
        )  # [N_atom, N_HKLs]
        dwf_aniso = DWF_aniso(
            atom_aniso_uw, orth2frac_tensor, sym_oped_hkl[:, i, :]
        )  # [N_atom, N_HKLs]
        dwf_all = torch.where(mask_vec[:, None], dwf_iso, dwf_aniso)
        exp_phase = exp_phase + dwf_all * torch.exp(1j * phase_G)
    F_calc = torch.sum(exp_phase * oc_sf, dim=0)
    return F_calc


def F_protein_batch(
    HKL_array,
    dr2_array,
    fullsf_tensor,
    R_G_tensor_stack,
    T_G_tensor_stack,
    orth2frac_tensor,
    atom_pos_frac_batch,
    atom_b_iso,
    atom_aniso_uw,
    atom_occ,
    PARTITION=20,
):
    """
    Calculate Protein Structural Factor from a batch of atomic models

    atom_pos_frac_batch: 3D tensor, [N_batch, N_atoms, N_dim=3]

    TODO: Support batched B factors
    """
    # F_calc = sum_Gsum_j{ [f0_sj*DWF*exp(2*pi*i*(h,k,l)*(R_G*(x1,x2,x3)+T_G))]} fractional postion, Rupp's Book P279
    # G is symmetry operations of the spacegroup and j is the atoms
    # DWF is the Debye-Waller Factor, has isotropic and anisotropic version, based on the PDB file input, Rupp's Book P641
    HKL_tensor = torch.tensor(HKL_array).to(R_G_tensor_stack)
    batchsize = atom_pos_frac_batch.shape[0]

    oc_sf = fullsf_tensor * atom_occ[..., None]  # [N_atom, N_HKLs]
    dwf_iso = DWF_iso(atom_b_iso, dr2_array)
    mask_vec = torch.all(torch.all(atom_aniso_uw == 0.0, dim=-1), dim=-1)
    # Vectorized phase calculation
    # sym_oped_pos_frac = (
    #     torch.tensordot(
    #         atom_pos_frac_batch, torch.permute(R_G_tensor_stack, [2, 1, 0]), 1
    #     )
    #     + T_G_tensor_stack.T
    # )
    # Shape [N_batch, N_atom, N_dim=3, N_ops]
    sym_oped_pos_frac = (
        torch.einsum("bay,oxy->baxo", atom_pos_frac_batch, R_G_tensor_stack)
        + T_G_tensor_stack.T
    )
    sym_oped_hkl = torch.einsum("rx,oxy->roy", HKL_tensor, R_G_tensor_stack)
    N_ops = R_G_tensor_stack.shape[0]
    N_partition = batchsize // PARTITION + 1
    F_calc = 0.0
    for j in range(N_partition):
        Fcalc_j = 0.0
        if j * PARTITION >= batchsize:
            continue
        start = j * PARTITION
        end = min((j + 1) * PARTITION, batchsize)
        for i in range(N_ops):  # Loop through symmetry operations to reduce memory cost
            # [N_atoms, N_HKLs]
            dwf_aniso = DWF_aniso(
                atom_aniso_uw, orth2frac_tensor, sym_oped_hkl[:, i, :]
            )  
            # [N_atoms, N_HKLs]
            dwf_all = torch.where(
                mask_vec[:, None], dwf_iso, dwf_aniso
            ) 
            # Shape [PARTITION, N_atoms, N_HKLs]
            exp_phase_ij = dwf_all * torch.exp(1j * (
                2
                * torch.pi
                * torch.tensordot(
                    sym_oped_pos_frac[start:end, :, :, i], HKL_tensor.T, 1
                )
            ))
            # Shape [PARTITION, N_HKLs], sum over atoms
            Fcalc_ij = torch.sum(exp_phase_ij * oc_sf, dim=1)
            del exp_phase_ij # release the memory
            # Shape [PARTITION, N_HKLs], sum over symmetry operations
            Fcalc_j = Fcalc_j + Fcalc_ij
            del Fcalc_ij
        if j == 0:
            F_calc = Fcalc_j
        else:
            # Shape [N_batches, N_HKLs]
            F_calc = torch.concat((F_calc, Fcalc_j), dim=0)  # type: ignore
    return F_calc



def getVersionNumber():
    import pkg_resources
    version = pkg_resources.require("SFcalculator_torch")[0].version
    return version

__author__ = "Minhuan Li"
__email__ = "minhuanli@g.harvard.edu"
__version__ = getVersionNumber()

# Top level API
from .Fmodel import SFcalculator
from .io import PDBParser, fetch_pdb, fetch_pdbredo
from .symmetry import get_polar_axis

# Submodules
from . import utils
from . import patterson
from . import packingscore





import torch
import numpy as np
from .voxel import voxelvalue_torch_p1


def packingscore_voxelgrid_torch(
    atom_pos_orth,
    unit_cell,
    space_group,
    vdw_rad,
    unitcell_grid_center_orth_tensor,
    CUTOFF=0.0001,
):
    """
    Calculate the grid occupancy and clash percentage as packing score

    Parameters
    ----------
    atom_pos_orth: torch.tensor, [N_atom, 3]
        asu model in cartesian coordinates, better remove all hydrogen atoms at the beginning

    unit_cell: gemmi.UnitCell

    space_group: gemmi.SpaceGroup

    vdw_rad: torch.tensor, [N_atom, ]
        vdw radius of atoms, use vdw_rad_tensor to calculate

    unitcell_grid_center_orth_tensor: torch.float32 tensor, [N_grid, 3]
        center positions of all grids in carteisian space, use unitcell_grid_center to calculate

    CUTOFF: float, default 0.0001, must < 0.5
        cutoff to convert into binary map; Larger cutoff means slower decay further

    Returns
    -------
    Percentage of the occupancy of all unitcell grids, and percentage
    of clash grids between all symmetrical pairs.
    """
    N_grid = len(unitcell_grid_center_orth_tensor[:, 0])
    spacing = torch.max(
        unitcell_grid_center_orth_tensor[1] - unitcell_grid_center_orth_tensor[0]
    )
    # s ~ log(1/c -1) / (d/2 - r)
    spacing = torch.maximum(spacing, torch.tensor(3.5))
    steepness = np.log(1.0 / CUTOFF - 1.0) / (spacing / 2.0 - 1.5)

    voxel_map_p1 = voxelvalue_torch_p1(
        unitcell_grid_center_orth_tensor,
        atom_pos_orth,
        unit_cell,
        space_group,
        vdw_rad,
        s=steepness,
        binary=True,
        cutoff=CUTOFF,
    )

    occupancy = torch.count_nonzero(voxel_map_p1) / N_grid
    clash = torch.count_nonzero(voxel_map_p1 > 1) / N_grid

    return occupancy, clash



from __future__ import annotations

import numpy as np
import gemmi
import torch
import reciprocalspaceship as rs
import pandas as pd
from typing import List

ccp4_hkl_asu = [
    0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3,
    3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 7, 6, 7, 6, 7, 7, 7,
    6, 7, 6, 7, 7, 6, 6, 7, 7, 7, 7, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9,
    9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9
]

asu_cases = {
    0: lambda h, k, l: (l > 0) | ((l == 0) & ((h > 0) | ((h == 0) & (k >= 0)))),
    1: lambda h, k, l: (k >= 0) & ((l > 0) | ((l == 0) & (h >= 0))),
    2: lambda h, k, l: (h >= 0) & (k >= 0) & (l >= 0),
    3: lambda h, k, l: (l >= 0) & (((h >= 0) & (k > 0)) | ((h == 0) & (k == 0))),
    4: lambda h, k, l: (h >= k) & (k >= 0) & (l >= 0),
    5: lambda h, k, l: ((h >= 0) & (k > 0)) | ((h == 0) & (k == 0) & (l >= 0)),
    6: lambda h, k, l: (h >= k) & (k >= 0) & ((k > 0) | (l >= 0)),
    7: lambda h, k, l: (h >= k) & (k >= 0) & ((h > k) | (l >= 0)),
    8: lambda h, k, l: (h >= 0) & (((l >= h) & (k > h)) | ((l == h) & (k == h))),
    9: lambda h, k, l: (k >= l) & (l >= h) & (h >= 0),
}

# Spagegroup with polar axis (arbitrary origin axis)
polar_cases = {
    1: [0, 1, 2], # P1
    3: [1], # P2
    4: [1], # P21
    5: [1], # C2
    75: [2], # P4
    76: [2], # P41
    77: [2], # P42
    78: [2], # P43
    79: [2], # I4
    80: [2], # I41
    143: [2], # P3
    144: [2], # P31
    145: [2], # P32
    146: [2], # R3;H
    168: [2], # P6
    169: [2], # P61
    170: [2], # P65
    171: [2], # P62
    172: [2], # P64
    173: [2], # P63
}

def expand_to_p1(spacegroup, Hasu_array, Fasu_tensor, Batch=False, dmin_mask=6.0, unitcell=None, anomalous=False):
    '''
    Expand the reciprocal ASU array to a complete p1 unit cell, with phase shift on the Complex Structure Factor
    In a fully differentiable manner (to Fasu_tensor), with tensorflow

    Parameters:
    -----------
    spacegroup: gemmi.SpaceGroup
        A gemmi spacegroup object

    Hasu_array: np.int32 array
        The HKL list of reciprocal ASU

    Fasu_tensor: torch.complex64 tensor, single model or batched model
        Corresponding structural factor tensor

    Batch: Boolean, Default False
        Use to show if a single or a batch of Fasu_tensor are given. If True, the Fasu_tensor 
        should be in shape [N_batch, N_Hasu]; If False, the Fasu_tensor should be in shape
        [N_Hasu,]

    dmin_mask: np.float32, Default 6 angstroms.
        Minimum resolution cutoff, in angstroms, for creating the solvent mask

    anomalous: Boolean, Default False
        If the anomalous scattering is included in the calculation

    Return:
    -------
    Hp1_array, Fp1_tensor
        HKL list in p1 unit cell and corresponding complex structural factor tensor
    '''

    groupops = spacegroup.operations()
    allops = [op for op in groupops]

    if Batch:
        # Batched calculation
        assert Fasu_tensor.dim() == 2, "Give batch Fasu if you set Batch=True!"
        concat_axis = 1
    else:
        # Single model calculation
        assert Fasu_tensor.dim() == 1, "Give single Fasu if you set Batch=False!"
        concat_axis = 0
    if dmin_mask is not None:
        # expands to p1 with resolution set by dmin_mask, to remove high-frequency noise.
        dHKL = unitcell.calculate_d_array(Hasu_array).astype("float32") # type: ignore
        new_hkl_inds_bool = (dHKL >= dmin_mask)
        Hasu_array = Hasu_array[new_hkl_inds_bool]

        # removes entries of Fasu_tensor that correspond to resolutions above dmin_mask
        if Batch:
            Fasu_tensor = Fasu_tensor[:,new_hkl_inds_bool]
        else: 
            Fasu_tensor = Fasu_tensor[new_hkl_inds_bool]

    Fp1_tensor = Fasu_tensor
    len_asu = len(Hasu_array)
    Hasu_tensor = torch.tensor(Hasu_array, device=Fasu_tensor.device).type(torch.float32)

    ds = pd.DataFrame()
    ds["H"] = Hasu_array[:, 0]
    ds["K"] = Hasu_array[:, 1]
    ds["L"] = Hasu_array[:, 2]
    ds["index"] = np.arange(len_asu)
    zero_tensor = torch.tensor(0.,device=Fasu_tensor.device)
    for i, op in enumerate(allops):
        if i == 0:
            continue
        rot_temp = np.array(op.rot)/op.DEN
        tran_temp = torch.tensor(np.array(op.tran)/op.DEN, device=Fasu_tensor.device).type(torch.float32)
        H_temp = np.matmul(Hasu_array, rot_temp).astype(np.int32)
        ds_temp = pd.DataFrame()
        ds_temp["H"] = H_temp[:, 0]
        ds_temp["K"] = H_temp[:, 1]
        ds_temp["L"] = H_temp[:, 2]
        ds_temp["index"] = np.arange(len_asu)+len_asu*i
        # exp(-2*pi*j*h*T)
        phaseshift_temp = torch.exp(torch.complex(
            zero_tensor, -2*np.pi*torch.tensordot(Hasu_tensor, tran_temp, dims=1)))
        Fcalc_temp = Fasu_tensor * phaseshift_temp
        ds = pd.concat([ds, ds_temp])
        Fp1_tensor = torch.concat((Fp1_tensor, Fcalc_temp), dim=concat_axis)

    if not anomalous:
        # Friedel Pair
        ds_friedel = ds.copy()
        ds_friedel["H"] = -ds_friedel["H"]
        ds_friedel["K"] = -ds_friedel["K"]
        ds_friedel["L"] = -ds_friedel["L"]
        ds_friedel["index"] = ds_friedel["index"] + len(ds)
        # TODO: Need to modify for anomalous scattering
        F_friedel_tensor = torch.conj(Fp1_tensor)

        # Combine
        ds = pd.concat([ds, ds_friedel])
        Fp1_tensor = torch.concat((Fp1_tensor, F_friedel_tensor), dim=concat_axis)

    ds = ds.drop_duplicates(subset=["H", "K", "L"])

    HKL_1 = ds[["H", "K", "L"]].values
    idx_1 = torch.tensor(ds["index"].values, device=Fp1_tensor.device)
    Fp1_tensor = torch.index_select(Fp1_tensor, dim=concat_axis, index=idx_1)
    in_asu = asu_cases[0]  # p1 symmetry
    idx_2 = in_asu(*HKL_1.T)
    HKL_2 = HKL_1[idx_2]
    idx_2i = torch.where(torch.tensor(idx_2, device=Fp1_tensor.device))[0]
    Fp1_tensor = torch.index_select(Fp1_tensor, dim=concat_axis, index=idx_2i)

    return HKL_2, Fp1_tensor


def generate_reciprocal_asu(cell, spacegroup, dmin, anomalous=False):
    """
    Generate the Miller indices of the reflections in the reciprocal ASU.
    If `anomalous=True` the Miller indices of acentric reflections will be
    included in both the Friedel-plus and Friedel-minus halves of reciprocal
    space. Centric Miller indices will only be included in the Friedel-plus
    reciprocal ASU.
    Parameters
    ----------
    cell : gemmi.UnitCell
        UnitCell object
    spacegroup : str, int, gemmi.SpaceGroup
        Space group to identify asymmetric unit
    dmin : float
        Maximum resolution of the data in Å
    anomalous : bool
        Whether to include Friedel-minus Miller indices to represent anomalous data
    Returns
    -------
    hasu : np.array (np.int32)
        n by 3 array of miller indices in the reciprocal ASU.
    """
    p1_hkl = generate_reciprocal_cell(cell, dmin)
    # Remove absences
    hkl = p1_hkl[~rs.utils.is_absent(p1_hkl, spacegroup)] #type: ignore
    # Map to ASU
    hasu = hkl[rs.utils.in_asu(hkl, spacegroup)] #type: ignore
    if anomalous:
        hasu_minus = -hasu[~rs.utils.is_centric(hasu, spacegroup)] #type: ignore
        return np.unique(np.concatenate([hasu, hasu_minus]), axis=0)
    return np.unique(hasu, axis=0)


def generate_reciprocal_cell(cell, dmin, dtype=np.int32):
    """
    Generate the miller indices of the full P1 reciprocal cell.
    Parameters
    ----------
    cell : gemmi.UnitCell
        Unit cell object
    dmin : float
        Maximum resolution of the data in Å
    dtype : np.dtype (optional)
        The data type of the returned array. The default is np.int32.
    Returns
    -------
    hkl : np.array(int32)
    """
    hmax, kmax, lmax = cell.get_hkl_limits(dmin)
    hkl = np.meshgrid(
        np.linspace(-hmax, hmax + 1, 2 * hmax + 2, dtype=dtype),
        np.linspace(-kmax, kmax + 1, 2 * kmax + 2, dtype=dtype),
        np.linspace(-lmax, lmax + 1, 2 * lmax + 2, dtype=dtype),
    )
    hkl = np.stack(hkl).reshape((3, -1)).T

    # Remove reflection 0,0,0
    hkl = hkl[np.any(hkl != 0, axis=1)]

    # Remove reflections outside of resolution range
    dHKL = cell.calculate_d_array(hkl).astype("float32")
    hkl = hkl[dHKL >= dmin]

    return hkl


def asu2p1_torch(atom_pos_orth, unitcell, spacegroup,
              incell=True, fractional=True):
    '''
    Apply symmetry operations to real space asu model coordinates

    Parameters
    ----------
    atom_pos_orth: tensor, [N_atom, 3]
        ASU model ccordinates
    
    unitcell: gemmi.UnitCell

    spacegroup: gemmi.SpaceGroup

    incell: boolean, default True
        If True, will move all atoms inside of the unit cell

    fractional: boolean, default True
        If True, will return fractional coordinates; Otherwise will return orthogonal coordinates

    Return
    ------
    atom_pos_sym_oped, [N_atoms, N_ops, 3] tensor in either fractional or orthogonal coordinates
    '''
    orth2frac_tensor = torch.tensor(unitcell.fractionalization_matrix.tolist(), device=atom_pos_orth.device, dtype=torch.float32)
    frac2orth_tensor = torch.tensor(unitcell.orthogonalization_matrix.tolist(), device=atom_pos_orth.device, dtype=torch.float32)
    operations = spacegroup.operations()  # gemmi.GroupOps object
    R_G_tensor_stack = torch.tensor(np.array([
            np.array(sym_op.rot)/sym_op.DEN for sym_op in operations]), device=atom_pos_orth.device).type(torch.float32)
    T_G_tensor_stack = torch.tensor(np.array([
            np.array(sym_op.tran)/sym_op.DEN for sym_op in operations]), device=atom_pos_orth.device).type(torch.float32)
    atom_pos_frac = torch.tensordot(atom_pos_orth, orth2frac_tensor.T, 1)
    sym_oped_pos_frac = torch.permute(torch.tensordot(R_G_tensor_stack,
        atom_pos_frac.T, 1), [2, 0, 1]) + T_G_tensor_stack
    if incell:
        sym_oped_pos_frac = sym_oped_pos_frac - \
            torch.floor(sym_oped_pos_frac)
    if fractional:
        return sym_oped_pos_frac
    else:
        sym_oped_pos_orth = torch.tensordot(sym_oped_pos_frac, frac2orth_tensor.T, 1)
        return sym_oped_pos_orth


def get_polar_axis(spacegroup : gemmi.SpaceGroup) -> List[int] | None:
    """
    Return list of polar axis of a spacegroup

    Args:
        spacegroup : gemmi.SpaceGroup
    
    Return:
        List[int] | None
    """
    sg_id = spacegroup.number
    return polar_cases.get(sg_id)


import numpy as np
import torch

import reciprocalspaceship as rs

import SFC_Torch.patterson as pat
from SFC_Torch.utils import try_gpu

import pytest


def test_Puvw(data_mtz_exp):
    mtz = rs.read_mtz(data_mtz_exp)
    mtz.dropna(axis=0, subset=['FP'], inplace=True)
    
    Fh = mtz['FP'].to_numpy()
    Fh2 = Fh**2
    HKL_array = mtz.get_hkls()
    cell = mtz.cell
    volume = cell.volume

    uvw_frac = pat.uvw_array_frac(cell, 3.0, 5.0, step=0.3)
    Puvw_np = pat.P_uvw_np(uvw_frac, Fh2, HKL_array, volume)

    uvw_tensor = torch.tensor(uvw_frac, dtype=torch.float32, device=try_gpu())
    HKL_tensor = torch.tensor(HKL_array, device=try_gpu())
    Fh_tensor = torch.tensor(Fh, device=try_gpu())
    Fh2_tensor = torch.tensor(Fh2, device=try_gpu())
    Puvw_torch = pat.P_uvw_torch(uvw_tensor, Fh2_tensor, HKL_tensor, volume)

    Pu = pat.Patterson_torch(uvw_frac, Fh_tensor, HKL_array, volume, PARTITION=1000)

    assert np.all(np.isclose(Puvw_np, Puvw_torch.cpu().numpy(), rtol=1e-03))
    assert np.all(np.isclose(Puvw_np, Pu.cpu().numpy(), rtol=1e-03))

@pytest.mark.parametrize("sharpen", [True, False])
@pytest.mark.parametrize("remove_origin", [True, False])
def test_Puvw_batch(data_mtz_exp, sharpen, remove_origin):
    
    mtz = rs.read_mtz(data_mtz_exp)
    mtz.dropna(axis=0, subset=['FP'], inplace=True)
    
    Fh = mtz['FP'].to_numpy()
    Fh2 = Fh**2
    HKL_array = mtz.get_hkls()
    cell = mtz.cell
    volume = cell.volume

    uvw_frac = pat.uvw_array_frac(cell, 3.0, 5.0, step=0.3)
    uvw_tensor = torch.tensor(uvw_frac, dtype=torch.float32, device=try_gpu())
    HKL_tensor = torch.tensor(HKL_array, device=try_gpu())
    Fh_tensor = torch.tensor(Fh, device=try_gpu())
    Fh2_tensor = torch.tensor(Fh2, device=try_gpu())
    
    Fh_batch = Fh_tensor.unsqueeze(0).repeat(10,1)
    Fh2_batch = Fh2_tensor.unsqueeze(0).repeat(10,1)
    
    Puvw_batch = pat.P_uvw_torch_batch(uvw_tensor, Fh2_batch, HKL_tensor, volume)

    Pu = pat.Patterson_torch(uvw_frac, Fh_tensor, HKL_array, volume, PARTITION=1000, sharpen=sharpen, remove_origin=remove_origin)
    Pu_batch = pat.Patterson_torch_batch(uvw_frac, Fh_batch, HKL_array, volume, PARTITION_uvw=1000, PARTITION_batch=5, sharpen=sharpen, remove_origin=remove_origin)

    if not sharpen and not remove_origin:
        assert torch.all(torch.isclose(Pu_batch, Puvw_batch, rtol=1e-03))
    
    assert torch.all(torch.isclose(Pu, Pu_batch[1], rtol=1e-03))













import pytest
import tempfile
from os.path import exists

import gemmi
import numpy as np

from SFC_Torch.io import PDBParser, fetch_pdb


def test_setdata(data_pdb):
    a = PDBParser(data_pdb)

    # test set_position
    new_pos = a.atom_pos.copy()
    new_pos[10, 2] = 7.88
    a.set_positions(new_pos)
    assert a.atom_pos[10, 2] == 7.88

    # test set_biso
    new_biso = a.atom_b_iso.copy()
    new_biso[10] = 7.88
    a.set_biso(new_biso)
    assert a.atom_b_iso[10] == 7.88

    # test set_baniso
    new_baniso = a.atom_b_aniso.copy()
    new_baniso[10, 1, 2] = 7.88
    a.set_baniso(new_baniso)
    assert a.atom_b_aniso[10, 1, 2] == 7.88
    
    # test set_occ
    new_occ = a.atom_occ.copy()
    new_occ[10] = 7.88
    a.set_occ(new_occ)
    assert a.atom_occ[10] == 7.88

    # test set_spacegroup
    a.set_spacegroup('I 4')
    assert a.spacegroup.hm == 'I 4'
    a.set_spacegroup(gemmi.SpaceGroup('P 41 3 2'))
    assert a.spacegroup.hm == 'P 41 3 2'

    # test set_unitcell
    a.set_unitcell(gemmi.UnitCell(50, 50, 50, 90, 90, 90))
    assert a.cell.parameters == (50, 50, 50, 90, 90, 90)


def test_savePDB(data_pdb):
    a = PDBParser(data_pdb)
    with tempfile.NamedTemporaryFile(suffix=".pdb") as temp:
        a.savePDB(temp.name, include_header=True)
        assert exists(temp.name)


def test_readcif(data_pdb, data_cif):
    a = PDBParser(data_pdb)
    b = PDBParser(data_cif)
    assert a.cra_name == b.cra_name
    assert (a.atom_pos == b.atom_pos).all()
    assert a.spacegroup.hm == b.spacegroup.hm

def test_saveCIF(data_pdb):
    a = PDBParser(data_pdb)
    with tempfile.NamedTemporaryFile(suffix=".cif") as temp:
        a.saveCIF(temp.name, include_header=True)
        assert exists(temp.name)

@pytest.mark.parametrize("inplace", [True, False])
def test_selection(data_pdb, inplace):
    a = PDBParser(data_pdb)
    b = a.selection('CA:*', inplace=inplace)
    if inplace:
        assert b is None
        assert len(a.atom_pos) == 55
    else:
        assert b.cell == a.cell
        assert b.spacegroup.hm == a.spacegroup.hm
        assert len(b.atom_pos) == 55


@pytest.mark.parametrize("inplace", [True, False])
def test_fromatomslices(data_pdb, inplace):
    a = PDBParser(data_pdb)
    is_CA = ["CA" in i for i in a.cra_name]
    CA_index = np.where(is_CA)[0]

    b = a.from_atom_slices(CA_index, inplace=inplace)
    if inplace:
        assert b is None
        assert len(a.atom_pos) == 55
    else:
        assert b.cell == a.cell
        assert b.spacegroup.hm == a.spacegroup.hm
        assert len(b.atom_pos) == 55

def test_fetchpdb():
    df = fetch_pdb(['4lZt', '1cTS', "1dUr"], outpath='../dev/')
    assert df['code'].tolist() == ['4lzt', '1cts', '1dur']
    assert df['with_pdb'].tolist() == [1, 1, 1]
    assert df['with_mtz'].tolist() == [0, 0, 0]
    assert df['with_sfcif'].tolist() == [1, 0, 1]
    assert df['with_mmcif'].tolist() == [1, 1, 1]
    assert exists("../dev/model_pdbs/4lzt.pdb")
    assert exists("../dev/model_pdbs/1dur.pdb")
    assert exists("../dev/model_pdbs/1cts.pdb")    
    assert exists("../dev/model_mmcifs/1dur.cif")
    assert exists("../dev/model_sfcifs/1dur-sf.cif")


from __future__ import annotations

import gemmi
import urllib.request, os

import numpy as np
from tqdm import tqdm
import pandas as pd

from .utils import assert_numpy


def hier2array(structure):
    """
    Convert the hierachical gemmi.structure into arrays of info

    structure: gemmi.Structure
    """
    atom_name = []
    cra_name = []
    atom_pos = []
    atom_b_aniso = []
    atom_b_iso = []
    atom_occ = []
    atom_altloc = []
    res_id = []
    i = 0
    model = structure[i]  # gemmi.Model object
    while len(model) == 0:  # Some PDB file has the empty first model
        i += 1
        try:
            model = structure[i]
        except:
            raise ValueError("Can't read valid model from the input PDB file!")
    j = 0
    for chain in model:
        for res in chain:
            for atom in res:
                # A list of atom name like ['O','C','N','C', ...], [Nc]
                atom_name.append(atom.element.name)
                # A list of atom long name like ['A-GLU-5-OD', ...], [Nc]
                cra_name.append(
                    chain.name + "-" + str(j) + "-" + res.name + "-" + atom.name
                )
                # A list of atom's Positions in orthogonal space, [Nc,3]
                atom_pos.append(atom.pos.tolist())
                # A list of anisotropic B Factor matrix, [Nc,3,3]
                atom_b_aniso.append(atom.aniso.as_mat33().tolist())
                # A list of isotropic B Factor [B1,B2,...], [Nc]
                atom_b_iso.append(atom.b_iso)
                # A list of occupancy [P1,P2,....], [Nc]
                atom_occ.append(atom.occ)
                # A list of altloc label
                atom_altloc.append(atom.altloc)
                # A list of residue id
                res_id.append(res.seqid.num)
            j += 1
    atom_pos = np.array(atom_pos)
    atom_b_aniso = np.array(atom_b_aniso)
    atom_b_iso = np.array(atom_b_iso)
    atom_occ = np.array(atom_occ)
    return atom_pos, atom_b_aniso, atom_b_iso, atom_occ, atom_name, cra_name, atom_altloc, res_id


def array2hier(
    atom_pos, atom_b_aniso, atom_b_iso, atom_occ, atom_name, cra_name, atom_altloc, res_id
):
    new_model = gemmi.Model("SFC")
    for i in range(len(cra_name)):
        Chain_i, resnum_i, resname_i, atomname_i = cra_name[i].split("-")
        if i == 0:
            current_chain = gemmi.Chain(Chain_i)
            current_res = gemmi.Residue()
            current_res.name = resname_i
            current_res.seqid = gemmi.SeqId(str(res_id[i]))
        else:
            if res_id[i] != current_res.seqid.num:
                current_chain.add_residue(current_res.clone())
                current_res = gemmi.Residue()
                current_res.name = resname_i
                current_res.seqid = gemmi.SeqId(str(res_id[i]))
            if Chain_i != current_chain.name:
                new_model.add_chain(current_chain.clone())
                current_chain = gemmi.Chain(Chain_i)

        current_atom = gemmi.Atom()
        current_atom.name = atomname_i
        current_atom.element = gemmi.Element(atom_name[i])
        Ui = atom_b_aniso[i]
        current_atom.aniso = gemmi.SMat33f(*[Ui[0,0], Ui[1,1], Ui[2,2], Ui[0,1], Ui[0,2], Ui[1,2]])
        current_atom.b_iso = atom_b_iso[i]
        current_atom.pos = gemmi.Position(*atom_pos[i])
        current_atom.occ = atom_occ[i]
        current_atom.altloc = atom_altloc[i]

        current_res.add_atom(current_atom)

    current_chain.add_residue(current_res.clone())
    new_model.add_chain(current_chain.clone())

    new_structure = gemmi.Structure()
    new_structure.add_model(new_model)
    new_structure.setup_entities()
    return new_structure


class PDBParser(object):
    """
    Read in the pdb file, and save atom name, atom positions, atom Biso, atom Baniso, atom occupancy in array manner
    Suppport indexing and gemmi-syntax structure selection
    """

    def __init__(self, data):
        """
        Create an PDBparser object from pbdfile

        data: pdb file path, or gemmi.Structure
        """
        if isinstance(data, str):
            if data.endswith("pdb"):
                structure = gemmi.read_pdb(data)
            elif data.endswith("cif"):
                structure = gemmi.make_structure_from_block(gemmi.cif.read(data)[0])
        elif isinstance(data, gemmi.Structure):
            structure = data
        else:
            raise KeyError(
                "data should be path str to a pdb file or a gemmi.Structure object"
            )
        (
            self.atom_pos,
            self.atom_b_aniso,
            self.atom_b_iso,
            self.atom_occ,
            self.atom_name,
            self.cra_name,
            self.atom_altloc,
            self.res_id,
        ) = hier2array(structure)
        try:
            self.spacegroup = gemmi.SpaceGroup(structure.spacegroup_hm)
        except:
            self.spacegroup = gemmi.SpaceGroup("P 1")
            print("No valid spacegroup in the file, set as P 1", flush=True)
        self.cell = structure.cell

        # Save the pdb headers, exclude the CRYST1 line
        header = structure.make_pdb_headers().split("\n")
        not_cryst = ["CRYST1" not in i for i in header]
        self.pdb_header = [header[i] for i in range(len(header)) if not_cryst[i]]

    @property
    def sequence(self):
        """
        Get one code squence
        """
        seen = set()
        unique_cras = [x for x in self.cra_name if x not in seen and not seen.add(x)]
        CA_cras = [i for i in unique_cras if "CA" in i]
        sequence = list(map(lambda x: x.split('-')[2], CA_cras))
        sequence = "".join([gemmi.find_tabulated_residue(r).one_letter_code for r in sequence])
        return sequence

    def to_gemmi(self, include_header=True):
        """
        Convert the array data to gemmi.Structure
        """
        st = array2hier(
            self.atom_pos,
            self.atom_b_aniso,
            self.atom_b_iso,
            self.atom_occ,
            self.atom_name,
            self.cra_name,
            self.atom_altloc,
            self.res_id,
        )
        st.spacegroup_hm = self.spacegroup.hm
        st.cell = self.cell
        if include_header:
            # I cheat here
            # In gemmi there is no method to set the whole header
            # because gemmi will parse all info in pdb headers except REMARKS
            # and save in corresponding attributes
            # Here a new gemmi.Structure object will be empty in all the attributes
            # So I set the raw_remarks to be the whole header
            # Next time user parse the new pdb with gemmi, will have all the info again
            st.raw_remarks = self.pdb_header
        return st
    
    @property
    def atom_pos_frac(self):
        return self.orth2frac(self.atom_pos)
    
    @property
    def operations(self):
        return self.spacegroup.operations() 

    @property
    def R_G_stack(self):
        """
        [n_ops, 3, 3], np.ndarray
        """
        return np.array([np.array(sym_op.rot) / sym_op.DEN for sym_op in self.operations])
    
    @property
    def T_G_stack(self):
        """
        [n_ops, 3], np.ndarray
        """
        return np.array([np.array(sym_op.tran) / sym_op.DEN for sym_op in self.operations])
    
    def exp_sym(self, frac_pos: np.ndarray | None = None) -> np.ndarray:
        """
        Apply all symmetry operations to the fractional coordinates

        Args:
            frac_pos, np.ndarray, [n_points, 3]
                fractional coordinates of model in single ASU. 
                If not given, will use self.atom_pos_frac
        Returns:
            np.ndarray, [n_points, n_ops, 3]
                fractional coordinates of symmetry operated models
        """
        if frac_pos is None:
            frac_pos = self.atom_pos_frac
        sym_oped_frac_pos = np.einsum("oxy,ay->aox", self.R_G_stack, frac_pos) + self.T_G_stack
        return sym_oped_frac_pos

    def set_spacegroup(self, spacegroup):
        """
        spacegroup: H-M notation of spacegroup or gemmi.SpaceGroup
        """
        if isinstance(spacegroup, str):
            self.spacegroup = gemmi.SpaceGroup(spacegroup)
        elif isinstance(spacegroup, gemmi.SpaceGroup):
            self.spacegroup = spacegroup

    def set_unitcell(self, unitcell):
        """
        unitcell: gemmi.UnitCell
        """
        self.cell = unitcell

    def set_positions(self, positions):
        """
        Set the atom positions with an array

        positions: array-like, [Nc, 3]
        """
        assert len(positions) == len(self.atom_pos), "Different atom number!"
        assert len(positions[0]) == 3, "Provide 3D coordinates!"
        self.atom_pos = positions

    def set_biso(self, biso):
        """
        Set the B-factors with an array

        biso: array-like, [Nc,]
        """
        assert len(biso) == len(self.atom_b_iso), "Different atom number!"
        assert biso[0].size == 1, "Provide one biso per atom!"
        self.atom_b_iso = biso

    def set_baniso(self, baniso):
        """
        Set the Anisotropic B-factors with an array

        baniso: array-like, [Nc,3,3]
        """
        assert len(baniso) == len(self.atom_b_aniso), "Different atom number!"
        assert baniso[0].shape == (3,3), "Provide a 3*3 matrix per atom!"
        self.atom_b_aniso = baniso

    def set_occ(self, occ):
        """
        Set the occupancy with an array

        occ: array-like, [Nc,]
        """
        assert len(occ) == len(self.atom_occ), "Different atom number!"
        assert occ[0].size == 1, "Provide one occupancy per atom!"
        self.atom_occ = occ

    def selection(self, condition, inplace=False):
        """
        Do structure selection with gemmi syntax

        condition: str, gemmi selection syntax, see https://gemmi.readthedocs.io/en/latest/analysis.html#selections
        """
        selection = gemmi.Selection(condition)
        structure = self.to_gemmi(include_header=False)
        sele_st = selection.copy_structure_selection(structure)

        if inplace:
            (
                self.atom_pos,
                self.atom_b_aniso,
                self.atom_b_iso,
                self.atom_occ,
                self.atom_name,
                self.cra_name,
                self.atom_altloc,
                self.res_id,
            ) = hier2array(sele_st)
        else:
            new_parser = PDBParser(sele_st)
            new_parser.pdb_header = self.pdb_header
            return new_parser

    def from_atom_slices(self, atom_slices, inplace=False):
        """
        Do selection or order change with atom slices

        atom_slices: array_like, index of the atoms you are interested in
        """
        if inplace:
            self.atom_pos = self.atom_pos[atom_slices]
            self.atom_b_aniso = self.atom_b_aniso[atom_slices]
            self.atom_b_iso = self.atom_b_iso[atom_slices]
            self.atom_occ = self.atom_occ[atom_slices]
            self.atom_name = [self.atom_name[i] for i in atom_slices]
            self.cra_name = [self.cra_name[i] for i in atom_slices]
            self.atom_altloc = [self.atom_altloc[i] for i in atom_slices]
            self.res_id = [self.res_id[i] for i in atom_slices]
        else:
            st = array2hier(
                self.atom_pos[atom_slices],
                self.atom_b_aniso[atom_slices],
                self.atom_b_iso[atom_slices],
                self.atom_occ[atom_slices],
                [self.atom_name[i] for i in atom_slices],
                [self.cra_name[i] for i in atom_slices],
                [self.atom_altloc[i] for i in atom_slices],
                [self.res_id[i] for i in atom_slices],
            )
            st.spacegroup_hm = self.spacegroup.hm
            st.cell = self.cell
            new_parser = PDBParser(st)
            new_parser.pdb_header = self.pdb_header
            return new_parser
        
    def move2cell(self):
        """
        move the current model into the cell by shifting
        """
        frac_mat = np.array(self.cell.fractionalization_matrix.tolist())
        mean_positions_frac = np.dot(frac_mat, np.mean(assert_numpy(self.atom_pos), axis=0))
        shift_vec = np.dot(np.linalg.inv(frac_mat), mean_positions_frac % 1.0 - mean_positions_frac)
        self.set_positions(assert_numpy(self.atom_pos) + shift_vec)

    def orth2frac(self, orth_pos: np.ndarray) -> np.ndarray:
        """
        Convert orthogonal coordinates to fractional coordinates

        Args:
            orth_pos: np.ndarray, [n_points, ..., 3]
        
        Returns:
            frational coordinates, np.ndarray, [n_points, ..., 3]
        """
        orth2frac_mat = np.array(self.cell.fractionalization_matrix.tolist())
        frac_pos = np.einsum("n...x,yx->n...y", orth_pos, orth2frac_mat)
        return frac_pos
    
    def frac2orth(self, frac_pos: np.ndarray) -> np.ndarray:
        """
        Convert fractional coordinates to orthogonal coordinates

        Args:
            frac_pos: np.ndarray, [n_points, ..., 3]
        
        Returns:
            orthogonal coordinates, np.ndarray, [n_points, ..., 3]
        """
        frac2orth_mat = np.array(self.cell.orthogonalization_matrix.tolist())
        orth_pos = np.einsum("n...x,yx->n...y", frac_pos, frac2orth_mat)
        return orth_pos
    
    def savePDB(self, savefilename, include_header=True):
        structure = self.to_gemmi(include_header=include_header)
        structure.write_pdb(savefilename)
    
    def saveCIF(self, savefilename, include_header=True):
        structure = self.to_gemmi(include_header=include_header)
        structure.make_mmcif_block().write_file(savefilename)


def fetch_pdb(idlist, outpath):
    '''
    Fetch pdb and mtz files from Protein Data Bank, with static urllib

    Parameters
    ----------
    idlist : [str]
        List of PDB ids
    
    outpath : str

    Returns
    -------
    DataFrame of fetch stats

    pdb files will be saved at outpath/models/
    mtz files will be saved at outpath/reflections/
    Record csv file will be saved at outpath/fetchpdb.csv
    '''

    if len(idlist) > 1:
        sequence_path = os.path.join(outpath, 'sequences/')
        model_path = os.path.join(outpath, 'model_pdbs/')
        mmcif_path = os.path.join(outpath, 'model_mmcifs/')
        sfcif_path = os.path.join(outpath, 'model_sfcifs/')
        reflection_path = os.path.join(outpath, 'reflections/')
        for folder in [sequence_path, model_path, mmcif_path, sfcif_path, reflection_path]:
            if os.path.exists(folder):
                print(f"{folder:<80}" + f"{'already exists': >20}")
            else:
                os.makedirs(folder)
                print(f"{folder:<80}" + f"{'created': >20}")
    else:
        sequence_path = outpath
        model_path = outpath
        mmcif_path = outpath
        sfcif_path = outpath
        reflection_path = outpath
    
    codes = []
    with_sequence = []
    with_pdb = []
    with_mmcif = []
    with_sfcif = []
    with_mtz = []
    for pdb_code in tqdm(idlist):
        valid_code = pdb_code.lower()
        seqlink = "https://www.rcsb.org/fasta/entry/" + valid_code.upper()
        pdblink = "https://files.rcsb.org/download/" + valid_code.upper() + ".pdb"
        mmciflink = "https://files.rcsb.org/download/" + valid_code.upper() + ".cif"
        sfciflink = "https://files.rcsb.org/download/" + valid_code.upper() + "-sf.cif"
        mtzlink = "https://edmaps.rcsb.org/coefficients/" + valid_code + ".mtz"
        codes.append(valid_code)

        try:
            urllib.request.urlretrieve(seqlink, os.path.join(sequence_path, valid_code+".fasta"))
            with_sequence.append(1)
        except:
            with_sequence.append(0)

        try:
            urllib.request.urlretrieve(pdblink, os.path.join(model_path, valid_code+".pdb"))
            with_pdb.append(1)
        except:
            with_pdb.append(0)

        try:
            urllib.request.urlretrieve(mmciflink, os.path.join(mmcif_path, valid_code+".cif"))
            with_mmcif.append(1)
        except:
            with_mmcif.append(0)

        try:
            urllib.request.urlretrieve(sfciflink, os.path.join(sfcif_path, valid_code+"-sf.cif"))
            with_sfcif.append(1)
        except:
            with_sfcif.append(0)

        try:
            urllib.request.urlretrieve(mtzlink, os.path.join(reflection_path, valid_code+".mtz"))
            with_mtz.append(1)
        except:
            with_mtz.append(0)
    
    stat_df = pd.DataFrame({
        "code" : codes,
        "with_sequence" : with_sequence,
        "with_pdb" : with_pdb,
        "with_mmcif" : with_mmcif,
        "with_sfcif" : with_sfcif,
        "with_mtz" : with_mtz
    })
    stat_df.to_csv(os.path.join(outpath, "fetchpdb.csv"))
    return stat_df


def fetch_pdbredo(idlist, outpath):
    '''
    Fetch re-refined and rebuilt moldes PDB-REDO, with static urllib
    see https://pdb-redo.eu/download for a full list of entry description for future development

    Parameters
    ----------
    idlist : [str]
        List of PDB ids
    
    outpath : str

    Returns
    -------
    DataFrame of fetch stats

    files will be saved at outpath/pdbredo_db/<pdb-id>/xxx
    '''

    codes = []
    with_pdb = []
    with_cif = []
    with_mtz = []
    with_version = []
    for pdb_code in tqdm(idlist):
        
        valid_code = pdb_code.lower()
        parent_link = f"https://pdb-redo.eu/db/{valid_code}/"
        pdblink = parent_link + f"{valid_code}_final.pdb"
        ciflink = parent_link + f"{valid_code}_final.cif"
        mtzlink = parent_link + f"{valid_code}_final.mtz"
        versionlink = parent_link + "versions.json"
        codes.append(valid_code)
        
        temp_path = os.path.join(outpath, "pdbredo_db", valid_code)
        if not os.path.exists(temp_path):
            os.makedirs(temp_path)

        try:
            urllib.request.urlretrieve(pdblink, os.path.join(temp_path, f"{valid_code}_final.pdb"))
            with_pdb.append(1)
        except:
            with_pdb.append(0)

        try:
            urllib.request.urlretrieve(ciflink, os.path.join(temp_path, f"{valid_code}_final.cif"))
            with_cif.append(1)
        except:
            with_cif.append(0)

        try:
            urllib.request.urlretrieve(mtzlink, os.path.join(temp_path, f"{valid_code}_final.mtz"))
            with_mtz.append(1)
        except:
            with_mtz.append(0)

        try:
            urllib.request.urlretrieve(versionlink, os.path.join(temp_path, f"{valid_code}_versions.json"))
            with_version.append(1)
        except:
            with_version.append(0)
    
    stat_df = pd.DataFrame({
        "code" : codes,
        "with_pdb" : with_pdb,
        "with_cif" : with_cif,
        "with_mtz" : with_mtz,
        "with_version" : with_version,
    })
    stat_df.to_csv(os.path.join(outpath, "fetchpdbredo.csv"))
    return stat_df









import torch


def reciprocal_grid(Hp1_array, Fp1_tensor, gridsize, batchsize=None):
    """
    Construct a reciprocal grid in reciprocal unit cell with HKL array and Structural Factor tensor
    Fully differentiable with torch

    Parameters
    ----------
    Hp1_array: np.int32 array
        The HKL list in the p1 unit cell. Usually the output of expand_to_p1

    Fp1_tensor: torch.complex64 tensor
        Corresponding structural factor tensor. Usually the output of expand_to_p1

    gridsize: array-like, int
        The size of the grid you want to create, the only requirement is "big enough"

    Return:
    Reciprocal space unit cell grid, as a torch.complex64 tensor
    """
    grid = torch.zeros(gridsize, device=Fp1_tensor.device, dtype=torch.complex64)
    tuple_index = tuple(torch.tensor(Hp1_array.T, device=Fp1_tensor.device, dtype=int))  # type: ignore
    if batchsize is not None:
        for i in range(batchsize):
            Fp1_tensor_i = Fp1_tensor[i]
            grid_i = grid.clone()
            grid_i[tuple_index] = Fp1_tensor_i  # Reciprocal Grid of model i
            if i == 0:
                grid_batch = grid_i[None, ...]
            else:
                grid_batch = torch.concat((grid_batch, grid_i[None, ...]), dim=0)  # type: ignore
        return grid_batch  # type: ignore
    else:
        grid[tuple_index] = Fp1_tensor
        return grid


def rsgrid2realmask(rs_grid, solvent_percent=0.50, exponent=50.0, Batch=False):
    """
    Convert reciprocal space grid to real space solvent mask grid, in a
    fully differentiable way with torch

    Parameters:
    -----------
    rs_grid: torch.complex64 tensor
        Reciprocal space unit cell grid. Usually the output of reciprocal_grid

    solvent_percent: 0 - 1 float
        The approximate volume percentage of solvent in the system, to generate the cutoff

    exponent: int/float
        The scale used in sigmoid function, to make the distribution binary

    Return:
    -------
    torch.float32 tensor
    The solvent mask grid in real space, solvent voxels have value close to 1, while protein voxels have value close to 0.
    """
    real_grid = torch.real(torch.fft.fftn(rs_grid, dim=(-3, -2, -1)))
    real_grid_norm = (real_grid - torch.mean(real_grid)) / torch.std(real_grid)
    if Batch:
        CUTOFF = torch.quantile(real_grid_norm[0], solvent_percent)
    else:
        CUTOFF = torch.quantile(real_grid_norm, solvent_percent)
    real_grid_mask = torch.sigmoid((CUTOFF - real_grid_norm) * exponent)
    return real_grid_mask


def realmask2Fmask(real_grid_mask, H_array, batchsize=None):
    """
    Convert real space solvent mask grid to mask structural factor, in a fully differentiable
    manner

    Parameters:
    -----------
    real_grid_mask: torch.float32 tensor
        The solvent massk grid in real space unit cell. Usually the output of rsgrid2realmask

    H_array: array-like, int
        The HKL list we are interested in to assign structural factors

    Return:
    -------
    torch.complex64 tensor
    Solvent mask structural factor corresponding to the HKL list in H_array
    """
    Fmask_grid = torch.fft.ifftn(real_grid_mask, dim=(-3, -2, -1), norm="forward")
    tuple_index = tuple(torch.tensor(H_array.T, device=Fmask_grid.device, dtype=int))  # type: ignore
    if batchsize is not None:
        Fmask = Fmask_grid[(slice(None), *tuple_index)]
    else:
        Fmask = Fmask_grid[tuple_index]
    return Fmask



import torch
import numpy as np


def uvw_array_orth(dmin, dmax, step=0.3):
    """
    create distance vector array within dmin to dmax range in orthogonal space
    usually choose dmin=3 and dmax=10, in unit of Angstrom
    """
    u_list = np.arange(0.0, dmax + 0.1, step)
    v_list = np.arange(0.0, dmax + 0.1, step)
    w_list = np.arange(0.0, dmax + 0.1, step)
    uvw_arr_orth = np.array(np.meshgrid(u_list, v_list, w_list)).T.reshape(-1, 3)
    d_array = np.sqrt(np.sum(uvw_arr_orth**2, axis=1))
    w = np.argwhere((d_array >= dmin) & (d_array <= dmax)).reshape(-1)
    return uvw_arr_orth[w]


# TODO: Return the ASU in the patterson space?


def uvw_array_frac(unitcell, dmin, dmax, step=0.3):
    """
    create distance vector array within dmin to dmax range in unitcell fractional space
    usually choose dmin=3 and dmax=10, in unit of Angstrom

    unitcell: Gemmi.UnitCell instance
    """
    uvw_arr_orth = uvw_array_orth(dmin, dmax, step=step)
    o2f_matrix = np.array(unitcell.fractionalization_matrix.tolist())
    return np.dot(uvw_arr_orth, o2f_matrix.T)


def P_uvw_np(uvw, Fh2, HKL_array, volume):
    """
    Compute the Patterson map for a given distance list uvw with numpy
    This is a purely vectorized function so will be memory-demanding
    so is designed for a small partition of full uvw list
    Typical length can be 10000, but depends on how large your memory is.

    volume: float, volume of real space unitcell, Gemmi.UnitCell.volume attribute
    """
    return (
        np.sum(
            Fh2
            * np.cos(
                2 * np.pi * np.sum(HKL_array[None, ...] * uvw[:, None, :], axis=-1)
            ),
            axis=-1,
        )
        * 2
        / volume
    )


def P_uvw_torch(uvw_frac_tensor, Fh2, HKL_tensor, volume):
    """
    Compute the Patterson map for a given distance list uvw with pytorch
    This is a purely vectorized function so will be memory-demanding
    so is designed for a small partition of full uvw list
    Typical length can be 10000, but can be much larger if you have a larger GPU
    60000 HKL with 10000 uvw will take ~0.9GB GPU memory

    Feed the function with tensors instead of arrays for a better performance
    """
    return (
        torch.sum(
            Fh2
            * torch.cos(
                2
                * np.pi
                * torch.sum(HKL_tensor[None, ...] * uvw_frac_tensor[:, None, :], dim=-1)
            ),
            dim=-1,
        )
        * 2
        / volume
    )


def P_uvw_torch_batch(uvw_frac_tensor, Fh2_batch, HKL_tensor, volume):
    """
    Similar to `P_uvw_torch` but now the Fh2_batch is a batch of models
    """
    return (
        torch.sum(
            Fh2_batch[:, None, :]
            * torch.cos(
                2
                * np.pi
                * torch.sum(HKL_tensor[None, ...] * uvw_frac_tensor[:, None, :], dim=-1)
            ),
            dim=-1,
        )
        * 2
        / volume
    )


def Patterson_torch(
    uvw_arr, Fh, HKL_array, volume, sharpen=False, remove_origin=False, PARTITION=10000
):
    """
    Differentiablly calculate the patterson map with pytorch

    Parameters
    ----------
    uvw_arr: array-like, with shape [N_uvw, 3]
        The distance vector array in fractional coordinates

    Fh: torch.tensor, with shape [N_hkl,]
        Structure factor amplitude

    HKL_array: array-like, with shape [N_hkl, 3]
        miller index array

    volume: float
        real space unitcell volume, Gemmi.UnitCell.volume attribute

    sharpen: bool, default False
        Whether normalize the SF magnitude to sharpen the patterson space.
        If True, do Eh = Fh/<Fh^2>^(1/2), Fh2_normed = (Eh^3 * Fh)^(1/2)
        Check Rupp's textbook P468 for details.

    remove_origin: bool, default False
        Whether remove the origin of the patterson map by subtract the Fh2.
        If True, do Fh2 = Fh^2 - <Fh^2>
        Check Rupp's textbook P468 for details.

    PARTITION: int, default 10000
        Size of partition for each vecotrized operation.
        Large partition size gives faster performance but more demanding to your memory.
        24000 HKL with 10000 parition requires about 0.9GB memory on GPU.
    """
    N_uvw = len(uvw_arr)
    N_partition = N_uvw // PARTITION + 1
    uvw_tensor = torch.tensor(uvw_arr).to(Fh)
    HKL_tensor = torch.tensor(HKL_array).to(Fh)
    Pu = 0.0

    if sharpen:
        Eh = Fh / torch.sqrt(torch.mean(Fh**2))
        Fh2 = torch.sqrt(Eh**3 * Fh)
    else:
        Fh2 = Fh**2

    if remove_origin:
        Fh2 = Fh2 - torch.mean(Fh2)

    for j in range(N_partition):
        Pu_j = 0.0
        if j * PARTITION >= N_uvw:
            continue
        start = j * PARTITION
        end = min((j + 1) * PARTITION, N_uvw)
        Pu_j = P_uvw_torch(uvw_tensor[start:end], Fh2, HKL_tensor, volume)

        if j == 0:
            Pu = Pu_j
        else:
            Pu = torch.concat([Pu, Pu_j], 0)
    return Pu


def Patterson_torch_batch(
    uvw_arr,
    Fh_batch,
    HKL_array,
    volume,
    sharpen=False,
    remove_origin=False,
    PARTITION_uvw=10000,
    PARTITION_batch=20,
    no_grad=False
):
    """
    Similar to `Patterson_torch` but for a batch of Fh

    Fh_batch: tf.tensor, [N_batch, N_hkl]

    PARTITION_uvw: int, default 10000
        Size of partition for each vecotrized operation.
        Large partition size gives faster performance but more demanding to your memory.
        60000 HKL with 10000 parition requires about 0.9GB memory on GPU.

    PARTITION_batch: int, default 20
        Size of partition on the batched model.
        24000 HKL with 10000 uvw partition and 20 models requires about 20GB memory on GPU
    
    no_grad: boolean, default False
        Turn off auto grad for batch calculation, save memory cost

    """
    N_uvw = len(uvw_arr)
    N_partition_uvw = N_uvw // PARTITION_uvw + 1

    N_batch = Fh_batch.shape[0]
    N_partition_batch = N_batch // PARTITION_batch + 1

    uvw_tensor = torch.tensor(uvw_arr).to(Fh_batch)
    HKL_tensor = torch.tensor(HKL_array).to(Fh_batch)
    Pu = 0.0

    if sharpen:
        Eh_batch = Fh_batch / torch.sqrt(
            torch.mean(Fh_batch**2, dim=-1, keepdims=True)
        )
        Fh2_batch = torch.sqrt(Eh_batch**3 * Fh_batch)
    else:
        Fh2_batch = Fh_batch**2

    if remove_origin:
        Fh2_batch = Fh2_batch - torch.mean(Fh2_batch, dim=-1, keepdims=True)

    for i in range(N_partition_batch):
        Pu_i = 0.0

        if i * PARTITION_batch >= N_batch:
            continue
        start_i = i * PARTITION_batch
        end_i = min((i + 1) * PARTITION_batch, N_batch)
        Fh2_i = Fh2_batch[start_i:end_i]

        for j in range(N_partition_uvw):
            Pu_ij = 0.0
            if j * PARTITION_uvw >= N_uvw:
                continue
            start_j = j * PARTITION_uvw
            end_j = min((j + 1) * PARTITION_uvw, N_uvw)
            Pu_ij = P_uvw_torch_batch(
                uvw_tensor[start_j:end_j], Fh2_i, HKL_tensor, volume
            )
            if no_grad:
                Pu_ij = Pu_ij.detach()

            if j == 0:
                Pu_i = Pu_ij
                del Pu_ij
            else:
                Pu_i = torch.concat([Pu_i, Pu_ij], 1)
                del Pu_ij

        if no_grad:
            Pu_i = Pu_i.detach()
        if i == 0:
            Pu = Pu_i
            del Pu_i
        else:
            Pu = torch.concat([Pu, Pu_i], 0)
            del Pu_i

    return Pu
